import os
from multiprocessing.pool import Pool

import discretize
import numpy as np
from scipy.sparse import csr_matrix as csr

from SimPEG.utils import mkvc

from ..simulation import LinearSimulation
from ..utils import validate_active_indices, validate_integer, validate_string

###############################################################################
#                                                                             #
#                             Base Potential Fields Simulation                #
#                                                                             #
###############################################################################


class BasePFSimulation(LinearSimulation):
    r"""Base class for potential field simulations that use integral formulations.

    For integral formulations, the forward simulation for a set of voxel cells
    can be defined as a linear operation of the form:

    .. math::
        \mathbf{d} = \mathbf{Am}

    where :math:`\mathbf{d}` are the data, :math:`\mathbf{m}` are the model paramters
    and :math:`\mathbf{A}` is a linear operator defining the sensitivities. The primary
    difference between child simulation classes is the kernel function used to create
    the rows of :math:`\mathbf{A}`.

    Parameters
    ----------
    mesh : discretize.TensorMesh or discretize.TreeMesh
        A 3D tensor or tree mesh.
    ind_active : np.ndarray of int or bool
        Indices array denoting the active topography cells.
    n_processes : None or int, optional
        The number of processes to use in the internal multiprocessing pool for forward
        modeling. The default value of 1 will not use multiprocessing. Any other setting
        will. `None` implies setting by the number of cpus.

    Notes
    -----
    If using multiprocessing by setting `n_processes` to a value other than 1, you must
    be aware of the method your operating system uses to spawn the subprocesses. On
    Windows the default method starts new processes that all import the main script.
    Therefor you must protect the calls to this class by testing if you are in
    the main process with:

    >>> from SimPEG.potential_fields import gravity
    >>> if __name__ == '__main__':
    ...     # Do your processing here
    ...     sim = gravity.Simulation3DIntegral(n_processes=4, ...)
    ...     sim.dpred(m)

    This usually does not affect jupyter notebook environments.
    """

    def __init__(
        self,
        mesh,
        ind_active=None,
        store_sensitivities="ram",
        n_processes=1,
        **kwargs,
    ):
        # If deprecated property set with kwargs
        if "actInd" in kwargs:
            raise AttributeError(
                "actInd was removed in SimPEG 0.17.0, please use ind_active"
            )

        if "forwardOnly" in kwargs:
            raise AttributeError(
                "forwardOnly was removed in SimPEG 0.17.0, please set store_sensitivities=None"
            )

        if "forward_only" in kwargs:
            raise AttributeError(
                "forward_only was removed in SimPEG 0.x.0, please set store_sensitivities=None"
            )

        super().__init__(mesh, store_sensitivities=store_sensitivities, **kwargs)
        self.solver = None
        self.n_processes = n_processes

        # Find non-zero cells indices
        if ind_active is None:
            ind_active = np.ones(mesh.n_cells, dtype=bool)
        else:
            ind_active = validate_active_indices("ind_active", ind_active, mesh.n_cells)
        self._ind_active = ind_active

        self.nC = int(sum(ind_active))

        if isinstance(mesh, discretize.TensorMesh):
            nodes = mesh.nodes
            inds = np.arange(mesh.n_nodes).reshape(mesh.shape_nodes, order="F")
            if mesh.dim == 2:
                cell_nodes = [
                    inds[:-1, :-1].reshape(-1, order="F"),
                    inds[1:, :-1].reshape(-1, order="F"),
                    inds[:-1, 1:].reshape(-1, order="F"),
                    inds[1:, 1:].reshape(-1, order="F"),
                ]
            if mesh.dim == 3:
                cell_nodes = [
                    inds[:-1, :-1, :-1].reshape(-1, order="F"),
                    inds[1:, :-1, :-1].reshape(-1, order="F"),
                    inds[:-1, 1:, :-1].reshape(-1, order="F"),
                    inds[1:, 1:, :-1].reshape(-1, order="F"),
                    inds[:-1, :-1, 1:].reshape(-1, order="F"),
                    inds[1:, :-1, 1:].reshape(-1, order="F"),
                    inds[:-1, 1:, 1:].reshape(-1, order="F"),
                    inds[1:, 1:, 1:].reshape(-1, order="F"),
                ]
            cell_nodes = np.stack(cell_nodes, axis=-1)[ind_active]
        elif isinstance(mesh, discretize.TreeMesh):
            nodes = np.r_[mesh.nodes, mesh.hanging_nodes]
            cell_nodes = mesh.cell_nodes[ind_active]
        else:
            raise ValueError("Mesh must be 3D tensor or Octree.")
        unique, unique_inv = np.unique(cell_nodes.T, return_inverse=True)
        self._nodes = nodes[unique]  # unique active nodes
        self._unique_inv = unique_inv.reshape(cell_nodes.T.shape)

    @property
    def n_processes(self):
        return self._n_processes

    @n_processes.setter
    def n_processes(self, value):
        if value is not None:
            value = validate_integer("n_processes", value, min_val=1)
        self._n_processes = value

    @property
    def ind_active(self):
        """Active topography cells.

        Returns
        -------
        (n_cell) numpy.ndarray of bool
            Returns the active topography cells
        """
        return self._ind_active

    @property
    def actInd(self):
        """'actInd' is deprecated. Use 'ind_active' instead."""
        raise AttributeError(
            "The 'actInd' property has been deprecated. "
            "Please use 'ind_active'. This will be removed in version 0.17.0 of SimPEG.",
        )

    def linear_operator(self):
        """Return linear operator.

        Returns
        -------
        numpy.ndarray
            Linear operator
        """
        n_cells = self.nC
        if getattr(self, "model_type", None) == "vector":
            n_cells *= 3
        if self.store_sensitivities == "disk":
            sens_name = os.path.join(self.sensitivity_path, "sensitivity.npy")
            if os.path.exists(sens_name):
                # do not pull array completely into ram, just need to check the size
                kernel = np.load(sens_name, mmap_mode="r")
                if kernel.shape == (self.survey.nD, n_cells):
                    print(f"Found sensitivity file at {sens_name} with expected shape")
                    kernel = np.asarray(kernel)
                    return kernel
        if self.n_processes == 1:
            kernel = []
            for args in self.survey._location_component_iterator():
                kernel.append(self.evaluate_integral(*args))
        else:
            # multiprocessed
            with Pool(processes=self.n_processes) as pool:
                kernel = pool.starmap(
                    self.evaluate_integral, self.survey._location_component_iterator()
                )
        if self.store_sensitivities is not None:
            kernel = np.vstack(kernel)
        else:
            kernel = np.concatenate(kernel)
        if self.store_sensitivities == "disk":
            print(f"writing sensitivity to {sens_name}")
            os.makedirs(self.sensitivity_path, exist_ok=True)
            np.save(sens_name, kernel)
        return kernel

    @property
    def Jmatrix(self):
        """
        Gravity forward operator
        """
        if getattr(self, "_Jmatrix", None) is None:
            self._Jmatrix = self.linear_operator()

        return self._Jmatrix


class BaseEquivalentSourceLayerSimulation(BasePFSimulation):
    """Base equivalent source layer simulation class.

    Parameters
    ----------
    mesh : discretize.BaseMesh
        A 2D tensor or tree mesh defining discretization along the x and y directions
    cell_z_top : numpy.ndarray or float
        Define the elevations for the top face of all cells in the layer. If an array,
        it should be the same size as the active cell set.
    cell_z_bottom : numpy.ndarray or float
        Define the elevations for the bottom face of all cells in the layer. If an array,
        it should be the same size as the active cell set.
    """

    def __init__(self, mesh, cell_z_top, cell_z_bottom, **kwargs):
        if mesh.dim != 2:
            raise AttributeError("Mesh to equivalent source layer must be 2D.")

        super().__init__(mesh, **kwargs)

        if isinstance(cell_z_top, (int, float)):
            cell_z_top = float(cell_z_top) * np.ones(self.nC)

        if isinstance(cell_z_bottom, (int, float)):
            cell_z_bottom = float(cell_z_bottom) * np.ones(self.nC)

        if (self.nC != len(cell_z_top)) | (self.nC != len(cell_z_bottom)):
            raise AttributeError(
                "'cell_z_top' and 'cell_z_bottom' must have length equal to number of",
                "cells, and match the number of active cells.",
            )

        all_nodes = self._nodes[self._unique_inv]
        all_nodes = [
            np.c_[all_nodes[0], cell_z_bottom],
            np.c_[all_nodes[1], cell_z_bottom],
            np.c_[all_nodes[2], cell_z_bottom],
            np.c_[all_nodes[3], cell_z_bottom],
            np.c_[all_nodes[0], cell_z_top],
            np.c_[all_nodes[1], cell_z_top],
            np.c_[all_nodes[2], cell_z_top],
            np.c_[all_nodes[3], cell_z_top],
        ]
        self._nodes = np.stack(all_nodes, axis=0)
        self._unique_inv = None


def progress(iteration, prog, final):
    """Progress (% complete) for constructing sensitivity matrix.

    Parameters
    ----------
    iteration : int
        Current rows
    prog : float
        Progress
    final : int
        Number of rows (= number of receivers)

    Returns
    -------
    float
        % completed
    """
    arg = np.floor(float(iteration) / float(final) * 10.0)

    if arg > prog:
        print("Done " + str(arg * 10) + " %")
        prog = arg

    return prog


def get_dist_wgt(mesh, receiver_locations, actv, R, R0):
    """Compute distance weights for potential field simulations.

    Parameters
    ----------
    mesh : discretize.BaseMesh
        A discretize mesh
    receiver_locations : (n, 3) numpy.ndarray
        Observation locations [x, y, z]
    actv : (n_cell) numpy.ndarray of bool
        Active cells vector [0:air , 1: ground]
    R : float
        Decay factor (mag=3, grav =2)
    R0 : float
        Stabilization factor. Usually a fraction of the minimum cell size

    Returns
    -------
    wr : (n_cell) numpy.ndarray
        Distance weighting model; 0 for all inactive cells
    """
    # Find non-zero cells
    if actv.dtype == "bool":
        inds = (
            np.asarray([inds for inds, elem in enumerate(actv, 1) if elem], dtype=int)
            - 1
        )
    else:
        inds = actv

    nC = len(inds)

    # Create active cell projector
    P = csr((np.ones(nC), (inds, range(nC))), shape=(mesh.nC, nC))

    # Geometrical constant
    p = 1 / np.sqrt(3)

    # Create cell center location
    Ym, Xm, Zm = np.meshgrid(
        mesh.cell_centers_y, mesh.cell_centers_x, mesh.cell_centers_z
    )
    hY, hX, hZ = np.meshgrid(mesh.h[1], mesh.h[0], mesh.h[2])

    # Remove air cells
    Xm = P.T * mkvc(Xm)
    Ym = P.T * mkvc(Ym)
    Zm = P.T * mkvc(Zm)

    hX = P.T * mkvc(hX)
    hY = P.T * mkvc(hY)
    hZ = P.T * mkvc(hZ)

    V = P.T * mkvc(mesh.cell_volumes)
    wr = np.zeros(nC)

    ndata = receiver_locations.shape[0]
    count = -1
    print("Begin calculation of distance weighting for R= " + str(R))

    for dd in range(ndata):
        nx1 = (Xm - hX * p - receiver_locations[dd, 0]) ** 2
        nx2 = (Xm + hX * p - receiver_locations[dd, 0]) ** 2

        ny1 = (Ym - hY * p - receiver_locations[dd, 1]) ** 2
        ny2 = (Ym + hY * p - receiver_locations[dd, 1]) ** 2

        nz1 = (Zm - hZ * p - receiver_locations[dd, 2]) ** 2
        nz2 = (Zm + hZ * p - receiver_locations[dd, 2]) ** 2

        R1 = np.sqrt(nx1 + ny1 + nz1)
        R2 = np.sqrt(nx1 + ny1 + nz2)
        R3 = np.sqrt(nx2 + ny1 + nz1)
        R4 = np.sqrt(nx2 + ny1 + nz2)
        R5 = np.sqrt(nx1 + ny2 + nz1)
        R6 = np.sqrt(nx1 + ny2 + nz2)
        R7 = np.sqrt(nx2 + ny2 + nz1)
        R8 = np.sqrt(nx2 + ny2 + nz2)

        temp = (
            (R1 + R0) ** -R
            + (R2 + R0) ** -R
            + (R3 + R0) ** -R
            + (R4 + R0) ** -R
            + (R5 + R0) ** -R
            + (R6 + R0) ** -R
            + (R7 + R0) ** -R
            + (R8 + R0) ** -R
        )

        wr = wr + (V * temp / 8.0) ** 2.0

        count = progress(dd, count, ndata)

    wr = np.sqrt(wr) / V
    wr = mkvc(wr)
    wr = np.sqrt(wr / (np.max(wr)))

    print("Done 100% ...distance weighting completed!!\n")

    return wr



import numpy as np
import scipy.special as spec
from ...utils import validate_ndarray_with_shape, validate_float


from ...survey import BaseSrc
from .waveforms import BaseVRMWaveform

#########################################
# BASE VRM SOURCE CLASS
#########################################


class BaseSrcVRM(BaseSrc):
    """Base VRM source class

    Parameters
    ----------
    receiver_list : list of SimPEG.electromagnetics.viscous_remanent_magnetization.receivers.Point
        A list of VRM receivers
    location : (3) array_like
        Source location
    waveform : SimPEG.electromagnetics.viscous_remanent_magnetization.waveforms.BaseVRMWaveform
        A VRM waveform
    """

    def __init__(self, receiver_list, location=None, waveform=None, **kwargs):
        if not isinstance(waveform, BaseVRMWaveform):
            AttributeError(
                "Waveform must be an instance of a VRM waveform class: StepOff, SquarePulse or Arbitrary"
            )

        super(BaseSrcVRM, self).__init__(receiver_list, location, **kwargs)
        self.waveform = waveform

    @property
    def nRx(self):
        """Total number of receiver locations

        Returns
        -------
        int
            Number of receiver locations
        """
        return np.sum(
            np.array([np.shape(rx.locations)[0] for rx in self.receiver_list])
        )

    @property
    def vnRx(self):
        """Vector number of receiver locations

        Returns
        -------
        list of int
            Number of receivers per source
        """
        return np.array([np.shape(rx.locations)[0] for rx in self.receiver_list])


#########################################
# MAGNETIC DIPOLE VRM SOURCE CLASS
#########################################


class MagDipole(BaseSrcVRM):
    """Magnetic dipole source

    Parameters
    ----------
    receiver_list : list of SimPEG.electromagnetics.viscous_remanent_magnetization.receivers.Point
        VRM receivers
    location : (3) array_like
        source location
    moment : (3) array_like
        dipole moment (mx, my, mz)
    waveform : SimPEG.electromagnetics.viscous_remanent_magnetization.waveforms.BaseVRMWaveform
        VRM waveform
    """

    def __init__(self, receiver_list, location, moment, waveform, **kwargs):
        if len(location) != 3:
            raise ValueError(
                "Tx location (x,y,z) must be given as a column vector of length 3."
            )

        if len(moment) != 3:
            raise ValueError(
                "Dipole moment (mx,my,mz) must be given as a column vector of length 3."
            )

        super(MagDipole, self).__init__(receiver_list, location, waveform, **kwargs)

        self.moment = moment

    @property
    def moment(self):
        """Dipole moment (mx, my, mz)

        Returns
        -------
        (3) numpy.ndarray
            Dipole moment (mx, my, mz)
        """
        return self._moment

    @moment.setter
    def moment(self, val):
        self._moment = validate_ndarray_with_shape(
            "moment", val, shape=("*",), dtype=float
        )

    def getH0(self, xyz):
        """Compute inducing field at locations xyz

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            locations at which primary field components are computed

        Returns
        -------
        (n, 3) numpy.ndarray
            Primary magnetic field at all xyz locations; organized as columns [Hx0,Hy0,Hz0]
        """

        m = self.moment
        r0 = self.location

        r = np.sqrt(
            (xyz[:, 0] - r0[0]) ** 2
            + (xyz[:, 1] - r0[1]) ** 2
            + (xyz[:, 2] - r0[2]) ** 2
        )
        mdotr = (
            m[0] * (xyz[:, 0] - r0[0])
            + m[1] * (xyz[:, 1] - r0[1])
            + m[2] * (xyz[:, 2] - r0[2])
        )

        hx0 = (1 / (4 * np.pi)) * (
            3 * (xyz[:, 0] - r0[0]) * mdotr / r**5 - m[0] / r**3
        )
        hy0 = (1 / (4 * np.pi)) * (
            3 * (xyz[:, 1] - r0[1]) * mdotr / r**5 - m[1] / r**3
        )
        hz0 = (1 / (4 * np.pi)) * (
            3 * (xyz[:, 2] - r0[2]) * mdotr / r**5 - m[2] / r**3
        )

        return np.c_[hx0, hy0, hz0]

    def _getRefineFlags(self, xyzc, refinement_factor, refinement_distance):
        """
        This function finds the refinement factor to be assigned to each cell

        REQUIRED ARGUMENTS:

        xyzc -- Cell-center locations as NX3 array

        refinement_factor -- Refinement factors

        refinement_distance -- Refinement radii

        OUTPUTS:

        refFlag -- Vector of length N with the refinement factor for each cell

        """

        refFlag = np.zeros(np.shape(xyzc)[0], dtype=np.int)

        r = np.sqrt(
            (xyzc[:, 0] - self.location[0]) ** 2
            + (xyzc[:, 1] - self.location[1]) ** 2
            + (xyzc[:, 2] - self.location[2]) ** 2
        )

        for nn in range(0, refinement_factor):
            k = (r < refinement_distance[nn] + 1e-5) & (
                refFlag < refinement_factor - nn + 1
            )
            refFlag[k] = refinement_factor - nn

        return refFlag


#########################################
# CIRCULAR LOOP VRM SOURCE CLASS
#########################################


class CircLoop(BaseSrcVRM):
    """Circular loop source

    Parameters
    ----------
    receiver_list : list of SimPEG.electromagnetics.viscous_remanent_magnetization.receivers.Point
        VRM receivers
    location : (3) array_like
        source location
    radius : float
        loop radius
    orientation : (2) array_like
        Circular loop normal azimuth and declination
    Imax : float
        Maximum current amplitude
    waveform : SimPEG.electromagnetics.viscous_remanent_magnetization.waveforms.BaseVRMWaveform
        VRM waveform
    """

    def __init__(
        self, receiver_list, location, radius, orientation, Imax, waveform, **kwargs
    ):
        if len(location) != 3:
            raise ValueError(
                "Tx location (x,y,z) must be given as a column vector of length 3."
            )

        if len(orientation) != 2:
            raise ValueError(
                "Circular loop transmitter orientation orientation defined by two angles (theta, alpha)."
            )

        super(CircLoop, self).__init__(receiver_list, location, waveform, **kwargs)

        self.orientation = validate_ndarray_with_shape(
            "orientation", orientation, shape=(2,), dtype=float
        )
        self.radius = validate_float("radius", radius, min_val=0.0, inclusive_min=False)
        self.Imax = validate_float("Imax", Imax)

    def getH0(self, xyz):
        """Compute inducing field at locations xyz

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            locations at which primary field components are computed

        Returns
        -------
        (n, 3) numpy.ndarray
            Primary magnetic field at all xyz locations; organized as columns [Hx0,Hy0,Hz0]
        """

        r0 = self.location
        theta = self.orientation[0]  # Azimuthal
        alpha = self.orientation[1]  # Declination
        a = self.radius
        I = self.Imax

        # Rotate x,y,z into coordinate axis of transmitter loop
        rot_x = np.r_[
            np.c_[1, 0, 0],
            np.c_[0, np.cos(np.pi * theta / 180), -np.sin(np.pi * theta / 180)],
            np.c_[0, np.sin(np.pi * theta / 180), np.cos(np.pi * theta / 180)],
        ]  # CCW ROTATION OF THETA AROUND X-AXIS

        rot_z = np.r_[
            np.c_[np.cos(np.pi * alpha / 180), -np.sin(np.pi * alpha / 180), 0],
            np.c_[np.sin(np.pi * alpha / 180), np.cos(np.pi * alpha / 180), 0],
            np.c_[0, 0, 1],
        ]  # CCW ROTATION OF (90-ALPHA) ABOUT Z-AXIS

        rot_mat = np.dot(rot_x, rot_z)  # THE ORDER MATTERS

        x1p = np.dot(
            np.c_[xyz[:, 0] - r0[0], xyz[:, 1] - r0[1], xyz[:, 2] - r0[2]],
            rot_mat[0, :].T,
        )
        x2p = np.dot(
            np.c_[xyz[:, 0] - r0[0], xyz[:, 1] - r0[1], xyz[:, 2] - r0[2]],
            rot_mat[1, :].T,
        )
        x3p = np.dot(
            np.c_[xyz[:, 0] - r0[0], xyz[:, 1] - r0[1], xyz[:, 2] - r0[2]],
            rot_mat[2, :].T,
        )

        s = np.sqrt(x1p**2 + x2p**2) + 1e-10  # Radial distance
        k = 4 * a * s / (x3p**2 + (a + s) ** 2)

        hxp = (
            (x1p / s)
            * (x3p * I / (2 * np.pi * s * np.sqrt(x3p**2 + (a + s) ** 2)))
            * (
                ((a**2 + x3p**2 + s**2) / (x3p**2 + (s - a) ** 2))
                * spec.ellipe(k)
                - spec.ellipk(k)
            )
        )
        hyp = (
            (x2p / s)
            * (x3p * I / (2 * np.pi * s * np.sqrt(x3p**2 + (a + s) ** 2)))
            * (
                ((a**2 + x3p**2 + s**2) / (x3p**2 + (s - a) ** 2))
                * spec.ellipe(k)
                - spec.ellipk(k)
            )
        )
        hzp = (I / (2 * np.pi * np.sqrt(x3p**2 + (a + s) ** 2))) * (
            ((a**2 - x3p**2 - s**2) / (x3p**2 + (s - a) ** 2)) * spec.ellipe(k)
            + spec.ellipk(k)
        )

        # Rotate the other way to get back into original coordinates
        rot_mat_t = rot_mat.T
        hx0 = np.dot(np.c_[hxp, hyp, hzp], rot_mat_t[0, :].T)
        hy0 = np.dot(np.c_[hxp, hyp, hzp], rot_mat_t[1, :].T)
        hz0 = np.dot(np.c_[hxp, hyp, hzp], rot_mat_t[2, :].T)

        return np.c_[hx0, hy0, hz0]

    def _getRefineFlags(self, xyzc, refinement_factor, refinement_distance):
        """
        This function finds the refinement factor to be assigned to each cell

        REQUIRED ARGUMENTS:

        xyzc -- Cell-center locations as NX3 array

        refinement_factor -- Refinement factors

        refinement_distance -- Refinement radii

        OUTPUTS:

        refFlag -- Vector of length N with the refinement factor for each cell

        """

        refFlag = np.zeros(np.shape(xyzc)[0], dtype=np.int)

        r0 = self.location
        a = self.radius
        theta = self.orientation[0]  # Azimuthal
        alpha = self.orientation[1]  # Declination

        # Rotate x,y,z into coordinate axis of transmitter loop
        rot_x = np.r_[
            np.c_[1, 0, 0],
            np.c_[0, np.cos(np.pi * theta / 180), -np.sin(np.pi * theta / 180)],
            np.c_[0, np.sin(np.pi * theta / 180), np.cos(np.pi * theta / 180)],
        ]  # CCW ROTATION OF THETA AROUND X-AXIS

        rot_z = np.r_[
            np.c_[np.cos(np.pi * alpha / 180), -np.sin(np.pi * alpha / 180), 0],
            np.c_[np.sin(np.pi * alpha / 180), np.cos(np.pi * alpha / 180), 0],
            np.c_[0, 0, 1],
        ]  # CCW ROTATION OF (90-ALPHA) ABOUT Z-AXIS

        rot_mat = np.dot(rot_x, rot_z)  # THE ORDER MATTERS

        x1p = np.dot(
            np.c_[xyzc[:, 0] - r0[0], xyzc[:, 1] - r0[1], xyzc[:, 2] - r0[2]],
            rot_mat[0, :].T,
        )
        x2p = np.dot(
            np.c_[xyzc[:, 0] - r0[0], xyzc[:, 1] - r0[1], xyzc[:, 2] - r0[2]],
            rot_mat[1, :].T,
        )
        x3p = np.dot(
            np.c_[xyzc[:, 0] - r0[0], xyzc[:, 1] - r0[1], xyzc[:, 2] - r0[2]],
            rot_mat[2, :].T,
        )
        r = np.sqrt(x1p**2 + x2p**2 + x3p**2)
        cosA = np.sqrt(x1p**2 + x2p**2) / r
        d = np.sqrt(a**2 + r**2 - 2 * a * r * cosA)

        for nn in range(0, refinement_factor):
            k = (d < refinement_distance[nn] + 1e-3) & (
                refFlag < refinement_factor - nn + 1
            )
            refFlag[k] = refinement_factor - nn

        return refFlag


#########################################
# LINE CURRENT VRM SOURCE CLASS
#########################################


class LineCurrent(BaseSrcVRM):
    """Line current source.

    Parameters
    ----------
    receiver_list : list of SimPEG.electromagnetics.time_domain.receivers.BaseRx
        List of TDEM receivers
    location : (n, 3) numpy.ndarray
        Array defining the node locations for the wire path. For inductive sources,
        you must close the loop.
    Imax : float
        Maximum current amplitude
    waveform : SimPEG.electromagnetics.viscous_remanent_magnetization.waveforms.BaseVRMWaveform
        VRM waveform
    """

    def __init__(self, receiver_list, location, Imax, waveform, **kwargs):
        super(LineCurrent, self).__init__(receiver_list, location, waveform, **kwargs)

        self.Imax = validate_float("Imax", Imax)

    @property
    def location(self):
        """Line current nodes locations

        Returns
        -------
        (n, 3) numpy.ndarray
            Line current node locations.
        """
        return self._location

    @location.setter
    def location(self, loc):
        self._location = loc = validate_ndarray_with_shape(
            "location", loc, shape=("*", 3)
        )

    def getH0(self, xyz):
        """Compute inducing field at locations xyz

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            locations at which primary field components are computed

        Returns
        -------
        (n, 3) numpy.ndarray
            Primary magnetic field at all xyz locations; organized as columns [Hx0,Hy0,Hz0]
        """

        # TRANSMITTER NODES
        I = self.Imax
        tx_nodes = self.location
        x1tr = tx_nodes[:, 0]
        x2tr = tx_nodes[:, 1]
        x3tr = tx_nodes[:, 2]

        nLoc = np.shape(xyz)[0]
        nSeg = np.size(x1tr) - 1

        hx0 = np.zeros(nLoc)
        hy0 = np.zeros(nLoc)
        hz0 = np.zeros(nLoc)

        for pp in range(0, nSeg):
            # Wire ends for transmitter wire pp
            x1a = x1tr[pp]
            x2a = x2tr[pp]
            x3a = x3tr[pp]
            x1b = x1tr[pp + 1]
            x2b = x2tr[pp + 1]
            x3b = x3tr[pp + 1]

            # Vector Lengths between points
            vab = np.sqrt((x1b - x1a) ** 2 + (x2b - x2a) ** 2 + (x3b - x3a) ** 2)
            vap = np.sqrt(
                (xyz[:, 0] - x1a) ** 2 + (xyz[:, 1] - x2a) ** 2 + (xyz[:, 2] - x3a) ** 2
            )
            vbp = np.sqrt(
                (xyz[:, 0] - x1b) ** 2 + (xyz[:, 1] - x2b) ** 2 + (xyz[:, 2] - x3b) ** 2
            )

            # Cosines from cos()=<v1,v2>/(|v1||v2|)
            cos_alpha = (
                (xyz[:, 0] - x1a) * (x1b - x1a)
                + (xyz[:, 1] - x2a) * (x2b - x2a)
                + (xyz[:, 2] - x3a) * (x3b - x3a)
            ) / (vap * vab)
            cos_beta = (
                (xyz[:, 0] - x1b) * (x1a - x1b)
                + (xyz[:, 1] - x2b) * (x2a - x2b)
                + (xyz[:, 2] - x3b) * (x3a - x3b)
            ) / (vbp * vab)

            # Determining Radial Vector From Wire
            dot_temp = (
                (x1a - xyz[:, 0]) * (x1b - x1a)
                + (x2a - xyz[:, 1]) * (x2b - x2a)
                + (x3a - xyz[:, 2]) * (x3b - x3a)
            )

            rx1 = (x1a - xyz[:, 0]) - dot_temp * (x1b - x1a) / vab**2
            rx2 = (x2a - xyz[:, 1]) - dot_temp * (x2b - x2a) / vab**2
            rx3 = (x3a - xyz[:, 2]) - dot_temp * (x3b - x3a) / vab**2

            r = np.sqrt(rx1**2 + rx2**2 + rx3**2)

            phi = (cos_alpha + cos_beta) / r

            # I/4*pi in each direction
            ix1 = I * (x1b - x1a) / (4 * np.pi * vab)
            ix2 = I * (x2b - x2a) / (4 * np.pi * vab)
            ix3 = I * (x3b - x3a) / (4 * np.pi * vab)

            # Add contribution from wire pp into array
            hx0 = hx0 + phi * (-ix2 * rx3 + ix3 * rx2) / r
            hy0 = hy0 + phi * (ix1 * rx3 - ix3 * rx1) / r
            hz0 = hz0 + phi * (-ix1 * rx2 + ix2 * rx1) / r

        return np.c_[hx0, hy0, hz0]

    def _getRefineFlags(self, xyzc, refinement_factor, refinement_distance):
        """
        This function finds the refinement factor to be assigned to each cell

        REQUIRED ARGUMENTS:

        xyzc -- Cell-center locations as NX3 array

        refinement_factor -- Refinement factors

        refinement_distance -- Refinement radii

        OUTPUTS:

        refFlag -- Vector of length N with the refinement factor for each cell

        """

        ref_flag = np.zeros(np.shape(xyzc)[0], dtype=np.int)

        nSeg = np.shape(self.location)[0] - 1

        for tt in range(0, nSeg):
            ref_flag_tt = np.zeros(np.shape(xyzc)[0], dtype=np.int)
            tx0 = self.location[tt, :]
            tx1 = self.location[tt + 1, :]
            a = (tx1[0] - tx0[0]) ** 2 + (tx1[1] - tx0[1]) ** 2 + (tx1[2] - tx0[2]) ** 2
            b = (
                2 * (tx1[0] - tx0[0]) * (tx0[0] - xyzc[:, 0])
                + 2 * (tx1[1] - tx0[1]) * (tx0[1] - xyzc[:, 1])
                + 2 * (tx1[2] - tx0[2]) * (tx0[2] - xyzc[:, 2])
            )

            for nn in range(0, refinement_factor):
                d = refinement_distance[nn] + 1e-3
                c = (
                    (tx0[0] - xyzc[:, 0]) ** 2
                    + (tx0[1] - xyzc[:, 1]) ** 2
                    + (tx0[2] - xyzc[:, 2]) ** 2
                    - d**2
                )
                e = np.array(b**2 - 4 * a * c, dtype=np.complex)

                q_pos = (-b + np.sqrt(e)) / (2 * a)
                q_neg = (-b - np.sqrt(e)) / (2 * a)

                k_pos = (
                    (np.abs(np.imag(q_pos)) > 1e-12)
                    | ((np.real(q_pos) < 0.0) & (np.real(q_neg) < 0.0))
                    | ((np.real(q_pos) > 1.0) & (np.real(q_neg) > 1.0))
                )
                k_neg = (
                    (np.abs(np.imag(q_pos)) > 1e-12)
                    | ((np.real(q_pos) < 0.0) & (np.real(q_neg) < 0.0))
                    | ((np.real(q_pos) > 1.0) & (np.real(q_neg) > 1.0))
                    | (k_pos)
                )

                ind = (~k_pos) & (~k_neg) & (ref_flag_tt < refinement_factor + 1 - nn)
                ref_flag_tt[ind] = refinement_factor - nn

            ref_flag = np.maximum(ref_flag, ref_flag_tt)

        return ref_flag



"""
========================================================
Base SimPEG Classes (:mod:`SimPEG`)
========================================================
.. currentmodule:: SimPEG

SimPEG is built off of several base classes that define the
general structure of simulations and inversion operations.

Simulations
===========

Base Simulations
----------------

.. autosummary::
  :toctree: generated/

  simulation.BaseSimulation
  simulation.BaseTimeSimulation
  simulation.LinearSimulation
  simulation.ExponentialSinusoidSimulation
  base.BasePDESimulation
  base.BaseElectricalPDESimulation
  base.BaseMagneticPDESimulation

Base Surveys, Sources and Receivers
-----------------------------------
.. autosummary::
  :toctree: generated/

  survey.BaseRx
  survey.BaseTimeRx
  survey.BaseSrc
  survey.BaseSurvey
  survey.BaseTimeSurvey

Models
-----------------------------------
.. autosummary::
  :toctree: generated/

  models.Model
  props.PhysicalProperty
  props.Derivative
  props.Invertible
  props.Reciprocal
  props.HasModel

Data
----
.. autosummary::
  :toctree: generated/

  data.Data
  data.SyntheticData

Fields
------
.. autosummary::
  :toctree: generated/

  fields.Fields
  fields.TimeFields

Mappings
--------
.. autosummary::
  :toctree: generated/

  maps.BaseParametric
  maps.ChiMap
  maps.ComboMap
  maps.ComplexMap
  maps.ExpMap
  maps.LinearMap
  maps.IdentityMap
  maps.InjectActiveCells
  maps.MuRelative
  maps.LogMap
  maps.ParametricBlock
  maps.ParametricCircleMap
  maps.ParametricEllipsoid
  maps.ParametricLayer
  maps.ParametricPolyMap
  maps.Projection
  maps.ReciprocalMap
  maps.SphericalSystem
  maps.Surject2Dto3D
  maps.SurjectFull
  maps.SurjectUnits
  maps.SurjectVertical1D
  maps.Weighting
  maps.Wires

Inversions
==========

Objective Function Pieces
-------------------------
.. autosummary::
  :toctree: generated/

  objective_function.BaseObjectiveFunction
  objective_function.ComboObjectiveFunction
  objective_function.L2ObjectiveFunction
  data_misfit.BaseDataMisfit
  data_misfit.L2DataMisfit

Optimizations
-------------
.. autosummary::
  :toctree: generated/

  optimization.ProjectedGradient
  optimization.BFGS
  optimization.GaussNewton
  optimization.InexactGaussNewton
  optimization.SteepestDescent
  optimization.NewtonRoot
  optimization.ProjectedGNCG
  optimization.Minimize
  optimization.Remember
  optimization.IterationPrinters
  optimization.StoppingCriteria

Base inversion pieces
---------------------
.. autosummary::
  :toctree: generated/

  inverse_problem.BaseInvProblem
  inversion.BaseInversion

"""

# import discretize as Mesh
import discretize
from discretize import tests

from .data import Data, SyntheticData
from . import data_misfit
from . import directives
from . import maps
from . import models
from . import inverse_problem
from . import inversion
from . import regularization
from . import survey
from . import simulation

from . import utils
from .utils import mkvc
from .utils import Report
from .utils.solver_utils import (
    _checkAccuracy,
    SolverWrapD,
    SolverWrapI,
    Solver,
    SolverCG,
    SolverDiag,
    SolverLU,
    SolverBiCG,
)

__version__ = "0.19.0"
__author__ = "SimPEG Team"
__license__ = "MIT"
__copyright__ = "2013 - 2020, SimPEG Team, http://simpeg.xyz"



import numpy as np
import scipy.sparse as sp
from ..utils import validate_ndarray_with_shape

from .. import utils
from .base import BaseSimilarityMeasure


class LinearCorrespondence(BaseSimilarityMeasure):
    r"""
    The petrophysical linear constraint for joint inversions.

    ..math::

        \phi_c({\mathbf m}_{\mathbf1},{\mathbf m}_{\mathbf2})
        = \lambda\sum_{i=1}^M (k_1*m_1 + k_2*m_2 + k_3)

    Assuming that we are working with two models only.

    """

    def __init__(self, mesh, wire_map, coefficients=None, **kwargs):
        super().__init__(mesh, wire_map, **kwargs)
        if coefficients is None:
            coefficients = np.r_[1.0, -1.0, 0.0]
        self.coefficients = coefficients

    @property
    def coefficients(self):
        """coefficients for the linear relationship between parameters.

        Returns
        -------
        (3) numpy.ndarray of float
        """
        return self._coefficients

    @coefficients.setter
    def coefficients(self, value):
        self._coefficients = validate_ndarray_with_shape(
            "coefficients", value, shape=(3,)
        )

    def relation(self, model):
        """
        Computes the values of petrophysical linear relationship between two different
        geophysical models.

        The linear relationship is defined as:

        f(m1, m2)  = k1*m1 + k2*m2 + k3

        :param numpy.ndarray model: stacked array of individual models
                                    np.c_[model1, model2,...]

        :rtype: float
        :return: linearly related petrophysical values of two different models,
                  dimension: M by 1, :M number of model parameters.

        """
        m1, m2 = self.wire_map * model
        k1, k2, k3 = self.coefficients

        return k1 * m1 + k2 * m2 + k3

    def __call__(self, model):
        """
        Computes the sum of values of petrophysical linear relationship
        between two different geophysical models.

        :param numpy.ndarray model: stacked array of individual models
                                    np.c_[model1, model2,...]

        :rtype: float
        :return: a scalar value.
        """

        result = self.relation(model)
        return 0.5 * result.T @ result

    def deriv(self, model):
        """Computes the Jacobian of the coupling term.

        :param list of numpy.ndarray ind_models: [model1, model2,...]

        :rtype: numpy.ndarray
        :return: result: gradient of the coupling term with respect to model1, model2,
                 :dimension 2M by 1, :M number of model parameters.
        """
        k1, k2, k3 = self.coefficients
        r = self.relation(model)
        dc_dm1 = k1 * r
        dc_dm2 = k2 * r

        result = np.r_[dc_dm1, dc_dm2]

        return result

    def deriv2(self, model, v=None):
        """Computes the Hessian of the linear coupling term.

        :param list of numpy.ndarray ind_models: [model1, model2, ...]
        :param numpy.ndarray v: vector to be multiplied by Hessian
        :rtype: scipy.sparse.csr_matrix if v is None
                numpy.ndarray if v is not None
        :return Hessian matrix: | h11, h21 | :dimension 2M*2M.
                                |          |
                                | h12, h22 |
        """

        k1, k2, k3 = self.coefficients
        if v is not None:
            v1, v2 = self.wire_map * v
            p1 = k1**2 * v1 + k2 * k1 * v2
            p2 = k2 * k1 * v1 + k2**2 * v2
            return np.r_[p1, p2]
        else:
            n = self.regularization_mesh.nC
            A = utils.sdiag(np.ones(n) * (k1**2))
            B = utils.sdiag(np.ones(n) * (k2**2))
            C = utils.sdiag(np.ones(n) * (k1 * k2))
            return sp.bmat([[A, C], [C, B]], format="csr")



import numpy as np
import discretize

from ....utils import mkvc
from .solutions_1d import get1DEfields


def homo1DModelSource(mesh, freq, sigma_1d):
    """
    Function that calculates and return background fields

    :param discretize.base.BaseMesh mesh: Holds information on the discretization
    :param float freq: The frequency to solve at
    :param numpy.ndarray sigma_1d: Background model of conductivity to base the calculations on, 1d model.
    :rtype: numpy.ndarray
    :return: eBG_bp, E fields for the background model at both polarizations with shape (mesh.nE, 2).

    """

    # Get a 1d solution for a halfspace background
    if mesh.dim == 1:
        mesh1d = mesh
    else:
        mesh1d = discretize.TensorMesh([mesh.h[-1]], [mesh.x0[-1]])

    # Note: Everything is using e^iwt
    e0_1d = get1DEfields(mesh1d, sigma_1d, freq)
    if mesh.dim == 1:
        eBG_px = mkvc(e0_1d, 2)
        eBG_py = -mkvc(
            e0_1d, 2
        )  # added a minus to make the results in the correct quadrents.
    elif mesh.dim == 2:
        ex_px = np.zeros(mesh.vnEx, dtype=complex)
        ey_px = np.zeros((mesh.nEy, 1), dtype=complex)
        for i in np.arange(mesh.vnEx[0]):
            ex_px[i, :] = -e0_1d
        eBG_px = np.vstack((mkvc(ex_px, 2), ey_px))
        # Setup y (north) polarization (_py)
        ex_py = np.zeros((mesh.nEx, 1), dtype="complex128")
        ey_py = np.zeros(mesh.vnEy, dtype="complex128")
        # Assign the source to ey_py
        for i in np.arange(mesh.vnEy[0]):
            ey_py[i, :] = e0_1d
        # ey_py[1:-1, 1:-1, 1:-1] = 0
        eBG_py = np.vstack((ex_py, mkvc(ey_py, 2), ez_py))
    elif mesh.dim == 3:
        # us the z component of ex_grid as lookup for solution
        edges_u, inv_edges = np.unique(mesh.gridEx[:, -1], return_inverse=True)
        map_to_edge_u = np.where(np.isclose(mesh1d.gridN, edges_u[:, None], atol=0.0))[
            1
        ]
        ex_px = -e0_1d[map_to_edge_u][inv_edges]
        ey_px = np.zeros(mesh.nEy, dtype=complex)
        ez_px = np.zeros(mesh.nEz, dtype=complex)
        eBG_px = np.r_[ex_px, ey_px, ez_px][:, None]

        edges_u, inv_edges = np.unique(mesh.gridEy[:, -1], return_inverse=True)
        map_to_edge_u = np.where(np.isclose(mesh1d.gridN, edges_u[:, None], atol=0.0))[
            1
        ]
        ex_py = np.zeros(mesh.nEx, dtype=complex)
        ey_py = e0_1d[map_to_edge_u][inv_edges]
        ez_py = np.zeros(mesh.nEz, dtype=complex)
        eBG_py = np.r_[ex_py, ey_py, ez_py][:, None]

    # Return the electric fields
    eBG_bp = np.hstack((eBG_px, eBG_py))
    return eBG_bp


def analytic1DModelSource(mesh, freq, sigma_1d):
    """
    Function that calculates and return background fields

    :param discretize.base.BaseMesh mesh: Holds information on the discretization
    :param float freq: The frequency to solve at
    :param numpy.ndarray sigma_1d: Background model of conductivity to base the calculations on, 1d model.
    :rtype: numpy.ndarray
    :return: eBG_bp, E fields for the background model at both polarizations with shape (mesh.nE, 2).

    """
    from SimPEG.NSEM.Utils import getEHfields

    # Get a 1d solution for a halfspace background
    if mesh.dim == 1:
        mesh1d = mesh
    elif mesh.dim == 2:
        mesh1d = discretize.TensorMesh([mesh.h[1]], np.array([mesh.x0[1]]))
    elif mesh.dim == 3:
        mesh1d = discretize.TensorMesh([mesh.h[2]], np.array([mesh.x0[2]]))

    # # Note: Everything is using e^iwt
    Eu, Ed, _, _ = getEHfields(mesh1d, sigma_1d, freq, mesh.nodes_z)
    # Make the fields into a dictionary of location and the fields
    e0_1d = Eu + Ed
    E1dFieldDict = dict(zip(mesh.nodes_z, e0_1d))
    if mesh.dim == 1:
        eBG_px = mkvc(e0_1d, 2)
        eBG_py = -mkvc(
            e0_1d, 2
        )  # added a minus to make the results in the correct quadrents.
    elif mesh.dim == 2:
        ex_px = np.zeros(mesh.vnEx, dtype=complex)
        ey_px = np.zeros((mesh.nEy, 1), dtype=complex)
        for i in np.arange(mesh.vnEx[0]):
            ex_px[i, :] = -e0_1d
        eBG_px = np.vstack((mkvc(ex_px, 2), ey_px))
        # Setup y (north) polarization (_py)
        ex_py = np.zeros((mesh.nEx, 1), dtype="complex128")
        ey_py = np.zeros(mesh.vnEy, dtype="complex128")
        # Assign the source to ey_py
        for i in np.arange(mesh.vnEy[0]):
            ey_py[i, :] = e0_1d
        # ey_py[1:-1, 1:-1, 1:-1] = 0
        eBG_py = np.vstack((ex_py, mkvc(ey_py, 2), ez_py))
    elif mesh.dim == 3:
        # Setup x (east) polarization (_x)
        ex_px = -np.array([E1dFieldDict[i] for i in mesh.gridEx[:, 2]]).reshape(-1, 1)
        ey_px = np.zeros((mesh.nEy, 1), dtype=complex)
        ez_px = np.zeros((mesh.nEz, 1), dtype=complex)
        # Construct the full fields
        eBG_px = np.vstack((ex_px, ey_px, ez_px))
        # Setup y (north) polarization (_py)
        ex_py = np.zeros((mesh.nEx, 1), dtype="complex128")
        ey_py = np.array([E1dFieldDict[i] for i in mesh.gridEy[:, 2]]).reshape(-1, 1)
        ez_py = np.zeros((mesh.nEz, 1), dtype="complex128")
        # Construct the full fields
        eBG_py = np.vstack((ex_py, mkvc(ey_py, 2), ez_py))

    # Return the electric fields
    eBG_bp = np.hstack((eBG_px, eBG_py))
    return eBG_bp


# def homo3DModelSource(mesh, model, freq):
#     """
#         Function that estimates 1D analytic background fields from a 3D model.

#         :param Simpeg mesh object mesh: Holds information on the discretization
#         :param float freq: The frequency to solve at
#         :param np.array sigma_1d: Background model of conductivity to base the calculations on, 1d model.
#         :rtype: numpy.ndarray (mesh.nE, 2)
#         :return: eBG_bp, E fields for the background model at both polarizations.

#     """

#     if mesh.dim < 3:
#         raise IOError('Input mesh has to have 3 dimensions.')


#     # Get the locations
#     a = mesh.gridCC[:, 0:2].copy()
#     unixy = np.unique(a.view(a.dtype.descr * a.shape[1])).view(float).reshape(-1, 2)
#     uniz = np.unique(mesh.gridCC[:, 2])
#     # # Note: Everything is using e^iwt
#     # Need to loop thourgh the xy locations, assess the model and calculate the fields at the phusdo cell centers.
#     # Then interpolate the cc fields to the edges.

#     e0_1d = get1DEfields(mesh1d, sigma_1d, freq)

#     elif mesh.dim == 3:
#         # Setup x (east) polarization (_x)
#         ex_px = np.zeros(mesh.vnEx, dtype=complex)
#         ey_px = np.zeros((mesh.nEy, 1), dtype=complex)
#         ez_px = np.zeros((mesh.nEz, 1), dtype=complex)
#         # Assign the source to ex_x
#         for i in np.arange(mesh.vnEx[0]):
#             for j in np.arange(mesh.vnEx[1]):
#                 ex_px[i, j, :] = -e0_1d
#         eBG_px = np.vstack((mkvc(ex_px, 2), ey_px, ez_px))
#         # Setup y (north) polarization (_py)
#         ex_py = np.zeros((mesh.nEx, 1), dtype='complex128')
#         ey_py = np.zeros(mesh.vnEy, dtype='complex128')
#         ez_py = np.zeros((mesh.nEz, 1), dtype='complex128')
#         # Assign the source to ey_py
#         for i in np.arange(mesh.vnEy[0]):
#             for j in np.arange(mesh.vnEy[1]):
#                 ey_py[i, j, :] = e0_1d
#         # ey_py[1:-1, 1:-1, 1:-1] = 0
#         eBG_py = np.vstack((ex_py, mkvc(ey_py, 2), ez_py))

#     # Return the electric fields
#     eBG_bp = np.hstack((eBG_px, eBG_py))
#     return eBG_bp



import numpy as np
import scipy.sparse as sp

from .base import BaseSimilarityMeasure
from ..utils import validate_type
from ..utils.mat_utils import coterminal

###############################################################################
#                                                                             #
#                                Cross-Gradient                               #
#                                                                             #
###############################################################################


class CrossGradient(BaseSimilarityMeasure):
    r"""
    The cross-gradient constraint for joint inversions.

    ..math::
        \phi_c(\mathbf{m_1},\mathbf{m_2}) = \lambda \sum_{i=1}^{M} \|
        \nabla \mathbf{m_1}_i \times \nabla \mathbf{m_2}_i \|^2

    All methods assume that we are working with two models only.

    """

    def __init__(self, mesh, wire_map, approx_hessian=True, normalize=False, **kwargs):
        super().__init__(mesh, wire_map=wire_map, **kwargs)
        self.approx_hessian = approx_hessian
        self._units = ["metric", "metric"]
        self.normalize = normalize
        regmesh = self.regularization_mesh

        if regmesh.mesh.dim not in (2, 3):
            raise ValueError("Cross-Gradient is only defined for 2D or 3D")
        self._G = regmesh.cell_gradient
        self._Av = sp.diags(np.sqrt(regmesh.vol)) * regmesh.average_face_to_cell

    @property
    def approx_hessian(self):
        """whether to use the semi-positive definate approximation for the hessian.
        Returns
        -------
        bool
        """
        return self._approx_hessian

    @approx_hessian.setter
    def approx_hessian(self, value):
        self._approx_hessian = validate_type("approx_hessian", value, bool)

    def _model_gradients(self, models):
        """
        Compute gradient on faces
        """
        gradients = []
        for unit, wire in zip(self.units, self.wire_map):
            model = wire * models
            if unit == "radian":
                gradient = []
                components = "xyz" if self.regularization_mesh.dim == 3 else "xy"
                for comp in components:
                    distances = getattr(self.regularization_mesh, f"cell_distances_{comp}")
                    cell_grad = getattr(
                        self.regularization_mesh, f"cell_gradient_{comp}"
                    )
                    gradient.append(coterminal(cell_grad * model * distances) / distances)

                gradient = np.hstack(gradient) / np.pi
            else:
                gradient = self._G @ model

            gradients.append(gradient)

        return gradients

    def _calculate_gradient(self, model, normalized=False, rtol=1e-6):
        """
        Calculate the spatial gradients of the model using central difference.

        Concatenates gradient components into a single array.
        [[x_grad1, y_grad1, z_grad1],
         [x_grad2, y_grad2, z_grad2],
         [x_grad3, y_grad3, z_grad3],...]

        :param numpy.ndarray model: model

        :rtype: numpy.ndarray
        :return: gradient_vector: array where each row represents a model cell,
                 and each column represents a component of the gradient.

        """
        regmesh = self.regularization_mesh
        Avs = [regmesh.aveFx2CC, regmesh.aveFy2CC]
        if regmesh.dim == 3:
            Avs.append(regmesh.aveFz2CC)
        Av = sp.block_diag(Avs)

        # Compute the gradients and concatenate components.
        grad_models = self._model_gradients(model)

        gradients = []
        for gradient in grad_models:
            gradient = (Av @ (gradient)).reshape((-1, regmesh.dim), order="F")

            if normalized:
                norms = np.linalg.norm(gradient, axis=-1)
                ind = norms <= norms.max() * rtol
                norms[ind] = 1.0
                gradient /= norms[:, None]
                gradient[ind] = 0.0
                # set gradient to 0 if amplitude of gradient is extremely small
            gradients.append(gradient)

        return gradients

    def calculate_cross_gradient(self, model, normalized=False, rtol=1e-6):
        """
        Calculates the cross-gradients of the models at each cell center.

        Parameters
        ----------
        model : numpy.ndarray
            The input model, which will be automatically separated into the two
            parameters internally
        normalized : bool, optional
            Whether to normalize the gradient
        rtol : float, optional
            relative cuttoff for small gradients in the normalization

        Returns
        -------
        cross_grad : numpy.ndarray
            The norm of the cross gradient vector in each active cell.
        """
        # Compute the gradients and concatenate components.
        grad_m1, grad_m2 = self._calculate_gradient(model, normalized=normalized, rtol=rtol)

        # for each model cell, compute the cross product of the gradient vectors.
        cross_prod = np.cross(grad_m1, grad_m2)
        if self.regularization_mesh.dim == 3:
            cross_prod = np.linalg.norm(cross_prod, axis=-1)

        return cross_prod

    def __call__(self, model):
        r"""
        Computes the sum of all cross-gradient values at all cell centers.

        :param numpy.ndarray model: stacked array of individual models
                                    np.c_[model1, model2,...]
        :param bool normalized: returns value of normalized cross-gradient if True

        :rtype: float
        :returns: the computed value of the cross-gradient term.


        ..math::

            \phi_c(\mathbf{m_1},\mathbf{m_2})
            = \lambda \sum_{i=1}^{M} \|\nabla \mathbf{m_1}_i \times \nabla \mathbf{m_2}_i \|^2
            = \sum_{i=1}^{M} \|\nabla \mathbf{m_1}_i\|^2 \ast \|\nabla \mathbf{m_2}_i\|^2
                - (\nabla \mathbf{m_1}_i \cdot \nabla \mathbf{m_2}_i )^2
            = \|\phi_{cx}\|^2 + \|\phi_{cy}\|^2 + \|\phi_{cz}\|^2

        (optional strategy, not used in this script)

        """
        # m1, m2 = (wire * model for wire in self.wire_map)
        Av = self._Av
        G = self._G
        g_m1, g_m2 = self._model_gradients(model)

        return 0.5 * np.sum(
            (Av @ g_m1**2) * (Av @ g_m2**2) - (Av @ (g_m1 * g_m2)) ** 2
        )

    def deriv(self, model):
        """
        Computes the Jacobian of the cross-gradient.

        :param list of numpy.ndarray ind_models: [model1, model2,...]

        :rtype: numpy.ndarray
        :return: result: gradient of the cross-gradient with respect to model1, model2

        """
        Av = self._Av
        G = self._G
        g_m1, g_m2 = self._model_gradients(model)

        return self.wire_map_deriv.T * np.r_[
            (((Av @ g_m2**2) @ Av) * g_m1) @ G
            - (((Av @ (g_m1 * g_m2)) @ Av) * g_m2) @ G,
            (((Av @ g_m1**2) @ Av) * g_m2) @ G
            - (((Av @ (g_m1 * g_m2)) @ Av) * g_m1) @ G,
        ]

    def deriv2(self, model, v=None):
        """
        Computes the Hessian of the cross-gradient.

        :param list of numpy.ndarray ind_models: [model1, model2, ...]
        :param numpy.ndarray v: vector to be multiplied by Hessian

        :rtype: scipy.sparse.csr_matrix if v is None
                numpy.ndarray if v is not None
        :return Hessian matrix if v is None
                Hessian multiplied by vector if v is not No

        """
        Av = self._Av
        G = self._G
        g_m1, g_m2 = self._model_gradients(model)

        if v is None:
            A = (
                G.T
                @ (
                    sp.diags(Av.T @ (Av @ g_m2**2))
                    - sp.diags(g_m2) @ Av.T @ Av @ sp.diags(g_m2)
                )
                @ G
            )

            C = (
                G.T
                @ (
                    sp.diags(Av.T @ (Av @ g_m1**2))
                    - sp.diags(g_m1) @ Av.T @ Av @ sp.diags(g_m1)
                )
                @ G
            )

            B = None
            BT = None
            if not self.approx_hessian:
                # d_m1_d_m2
                B = (
                    G.T
                    @ (
                        2 * sp.diags(g_m1) @ Av.T @ Av @ sp.diags(g_m2)
                        - sp.diags(g_m2) @ Av.T @ Av @ sp.diags(g_m1)
                        - sp.diags(Av.T @ Av @ (g_m1 * g_m2))
                    )
                    @ G
                )
                BT = B.T

            return self.wire_map_deriv.T * sp.bmat([[A, B], [BT, C]], format="csr") * self.wire_map_deriv
        else:
            v1, v2 = (wire * v for wire in self.wire_map)

            Gv1 = G @ v1
            Gv2 = G @ v2

            p1 = G.T @ (
                (Av.T @ (Av @ g_m2**2)) * Gv1 - g_m2 * (Av.T @ (Av @ (g_m2 * Gv1)))
            )
            p2 = G.T @ (
                (Av.T @ (Av @ g_m1**2)) * Gv2 - g_m1 * (Av.T @ (Av @ (g_m1 * Gv2)))
            )

            if not self.approx_hessian:
                p1 += G.T @ (
                    2 * g_m1 * (Av.T @ (Av @ (g_m2 * Gv2)))
                    - g_m2 * (Av.T @ (Av @ (g_m1 * Gv2)))
                    - (Av.T @ (Av @ (g_m1 * g_m2))) * Gv2
                )

                p2 += G.T @ (
                    2 * g_m2 * (Av.T @ (Av @ (g_m1 * Gv1)))
                    - g_m1 * (Av.T @ (Av @ (g_m2 * Gv1)))
                    - (Av.T @ (Av @ (g_m2 * g_m1))) * Gv1
                )
            return self.wire_map_deriv.T * np.r_[p1, p2]



import numpy as np
from scipy.constants import mu_0, epsilon_0
from scipy import integrate


# useful params
def omega(frequency):
    """Compute angular frequency from frequency

    Parameters
    ----------
    frequency : float or numpy.ndarray
        Frequencies in Hz

    Returns
    -------
    float or numpy.ndarray
        Angular frequencies in rad/s
    """
    return 2.0 * np.pi * frequency


def k(frequency, sigma, mu=mu_0, eps=epsilon_0):
    r"""Wavenumber for EM waves in homogeneous media

    See eq 1.47 - 1.49 in Ward and Hohmann

    Parameters
    ----------
    frequency : float or numpy.ndarray
        Frequencies in Hz
    sigma : float
        Electrical conductivity in S/m
    mu : float, default: :math:`4\pi \;\times\; 10^{-7}` H/m
        Magnetic permeability in H/m
    eps : float, default: 8.8541878128 :math:`\times \; 1-^{-12}` F/m
        Dielectric permittivity in F/m

    Returns
    -------
    complex or numpy.ndarray
        Wavenumbers at all input frequencies
    """
    w = omega(frequency)
    alp = w * np.sqrt(mu * eps / 2 * (np.sqrt(1.0 + (sigma / (eps * w)) ** 2) + 1))
    beta = w * np.sqrt(mu * eps / 2 * (np.sqrt(1.0 + (sigma / (eps * w)) ** 2) - 1))
    return alp - 1j * beta


def VTEMFun(time, ta, tb, a):
    """VTEM waveform function

    Parameters
    ----------
    time : numpy.ndarray
        Times vector
    ta : float
        Time at peak exponential
    tb : float
        Start of off-time

    Returns
    -------
    (n_time) numpy.ndarray
        The waveform evaluated at all input times
    """
    out = np.zeros(time.size)
    out[time <= ta] = (1 - np.exp(-a * time[time <= ta] / ta)) / (1 - np.exp(-a))
    out[(time > ta) & (time < tb)] = (
        -1 / (tb - ta) * (time[(time > ta) & (time < tb)] - tb)
    )
    return out


def convolve_with_waveform(func, waveform, times, fargs=None, fkwargs=None):
    """convolves the function with the given waveform

    This convolves the given function with the waveform and evaluates it at the given
    times. This uses a high order Gaussian-Legendre quadrature to evaluate, and could
    likely be slow.

    Parameters
    ----------
    func : callable
        function of `t` that should be convolved
    waveform : SimPEG.electromagnetics.time_domain.waveforms.BaseWaveform
    times : array_like
    fargs : list, optional
        extra arguments given to `func`
    fkwargs : dict, optional
        keyword arguments given to `func`

    Returns
    -------
    numpy.ndarray
        the convolution evaluate at the given times
    """
    try:
        t_nodes = waveform.time_nodes
    except AttributeError:
        raise TypeError(f"Unsupported waveform type of {type(waveform)}")

    if fargs is None:
        fargs = []
    if fkwargs is None:
        fkwargs = {}

    n_int = len(t_nodes) - 1
    out = np.zeros_like(times, dtype=float)
    for it, t in enumerate(times):

        def integral(quad_time, t):
            wave_eval = waveform.eval_deriv(t - quad_time)
            return wave_eval * func(quad_time, *fargs, **fkwargs)

        for i in range(n_int):
            b = t - t_nodes[i]
            a = t - t_nodes[i + 1]
            # just do not evaluate the integral at negative times...
            a = np.maximum(a, 0.0)
            b = np.maximum(b, 0.0)
            val, _ = integrate.quadrature(integral, a, b, tol=0.0, maxiter=500, args=t)
            out[it] -= val
    return out



import numpy as np

from . import receivers
from . import sources
from .survey import Survey

from ..utils import *


def WennerSrcList(n_electrodes, a_spacing, in2D=False, plotIt=False):
    """
    Source list for a Wenner Array

    Parameters
    ----------
    n_electrodes : int
        Number of electrodes in the Wenner array
    a_spacing : float
        Wenner array spacing parameter *a*
    in2D : bool, default=``False``
        If ``True``, create 2D sources
    plotIt : bool, default=``False``
        If ``True``, plot the Wenner geometry

    Returns
    -------
    list of SimPEG.electromagnetics.static.resistivity.sources.Dipole
        List of sources and their associate receivers for the Wenner survey
    """

    elocs = np.arange(0, a_spacing * n_electrodes, a_spacing)
    elocs -= (n_electrodes * a_spacing - a_spacing) / 2
    space = 1
    WENNER = np.zeros((0,), dtype=int)
    for _ in range(n_electrodes):
        for jj in range(n_electrodes):
            test = np.r_[jj, jj + space, jj + space * 2, jj + space * 3]
            if np.any(test >= n_electrodes):
                break
            WENNER = np.r_[WENNER, test]
        space += 1
    WENNER = WENNER.reshape((-1, 4))

    if plotIt:
        for i, s in enumerate("rbkg"):
            plt.plot(elocs[WENNER[:, i]], s + ".")
        plt.show()

    # Create sources and receivers
    i = 0
    if in2D:

        def getLoc(ii, abmn):
            return np.r_[elocs[WENNER[ii, abmn]], 0]

    else:

        def getLoc(ii, abmn):
            return np.r_[elocs[WENNER[ii, abmn]], 0, 0]

    source_list = []
    for i in range(WENNER.shape[0]):
        rx = receivers.Dipole(
            getLoc(i, 1).reshape([1, -1]), getLoc(i, 2).reshape([1, -1])
        )
        src = sources.Dipole([rx], getLoc(i, 0), getLoc(i, 3))
        source_list += [src]

    return source_list


def _mini_pole_pole(survey, verbose=False):
    """Function to miniaturize a survey for use in DCSimulation.

    Miniaturizes the survey into the minimum number of unique pole-pole electrode
    combinations for AM, AN, BM, BN pairs (also taking into account reciprocity),
    for use in a DCSimulation only.
    """
    A = survey.locations_a
    B = survey.locations_b
    M = survey.locations_m
    N = survey.locations_n

    elecs, inverse = np.unique(np.r_[A, B, M, N], axis=0, return_inverse=True)

    inv_A, inv_B, inv_M, inv_N = inverse.reshape(4, -1)
    dipole_tx = inv_A != inv_B
    dipole_rx = inv_M != inv_N

    AM = np.sort(np.c_[inv_A, inv_M])
    AN = np.sort(np.c_[inv_A[dipole_rx], inv_N[dipole_rx]])
    BM = np.sort(np.c_[inv_B[dipole_tx], inv_M[dipole_tx]])
    BN = np.sort(np.c_[inv_B[dipole_tx & dipole_rx], inv_N[dipole_tx & dipole_rx]])
    unique_pole_poles, pole_pole_inv = np.unique(
        np.r_[AM, AN, BM, BN], axis=0, return_inverse=True
    )

    inv_AM, pole_pole_inv = pole_pole_inv[: len(AM)], pole_pole_inv[len(AM) :]
    inv_AN, pole_pole_inv = pole_pole_inv[: len(AN)], pole_pole_inv[len(AN) :]
    inv_BM, inv_BN = pole_pole_inv[: len(BM)], pole_pole_inv[len(BM) :]

    if verbose:
        print(f"There are {unique_pole_poles.shape[0]} unique pole-pole combinations.")

    unique_sources = []
    last_src = None
    i_d = 0
    while i_d < len(unique_pole_poles):
        if last_src != unique_pole_poles[i_d, 0]:
            last_src = unique_pole_poles[i_d, 0]
            rxs = []
        else:
            while (
                i_d < len(unique_pole_poles) and last_src == unique_pole_poles[i_d, 0]
            ):
                rxs.append(unique_pole_poles[i_d, 1])
                i_d += 1
            rxs = np.array(rxs, dtype=int)
            rxs = receivers.Pole(elecs[rxs])
            unique_sources.append(sources.Pole([rxs], elecs[last_src]))

    dipoles = [dipole_rx, dipole_tx]
    invs = [inv_AM, inv_AN, inv_BM, inv_BN]
    mini_survey = Survey(unique_sources)
    return dipoles, invs, mini_survey



"""
=======================================================================
Richards Flow (:mod:`SimPEG.flow.richards`)
=======================================================================
.. currentmodule:: SimPEG.flow.richards

About ``Richards flow``

Simulations
===========
.. autosummary::
  :toctree: generated/

  SimulationNDCellCentered

Survey, Sources and Receivers
=============================
.. autosummary::
  :toctree: generated/

  receivers.Pressure
  receivers.Saturation
  survey.Survey

Empirical utilities
===================
.. autosummary::
  :toctree: generated/

  empirical.NonLinearModel
  empirical.BaseWaterRetention
  empirical.BaseHydraulicConductivity
  empirical.Haverkamp_theta
  empirical.Haverkamp_k
  empirical.haverkamp
  empirical.HaverkampParams
  empirical.Vangenuchten_theta
  empirical.Vangenuchten_k
  empirical.van_genuchten
  empirical.VanGenuchtenParams

"""
from . import empirical
from .survey import Survey
from .simulation import SimulationNDCellCentered
from . import receivers

SimulationNDCellCentred = SimulationNDCellCentered



import numpy as np
import scipy.sparse as sp

from .base import BaseSimilarityMeasure


###############################################################################
#                                                                             #
#                            Joint Total Variation                            #
#                                                                             #
###############################################################################


class JointTotalVariation(BaseSimilarityMeasure):
    r"""
    The joint total variation constraint for joint inversions.

    ..math ::
        \phi_sim(\mathbf{m_1},\mathbf{m_2}) = \lambda \sum_{i=1}^{M} V_i\sqrt{|
        \nabla \mathbf{m_1}_i|^2 +|\nabla \mathbf{m_2}_i|^2}
    """

    def __init__(self, mesh, wire_map, eps=1e-8, **kwargs):
        super().__init__(mesh, wire_map=wire_map, **kwargs)
        self.set_weights(volume=self.regularization_mesh.vol)
        self.eps = eps

        self._G = self.regularization_mesh.cell_gradient

    @property
    def W(self):
        """
        Weighting matrix
        """
        if getattr(self, "_W", None) is None:
            weights = np.prod(list(self._weights.values()), axis=0)
            self._W = (
                sp.diags(weights**2) * self.regularization_mesh.average_face_to_cell
            )
        return self._W

    @property
    def wire_map(self):
        return self._wire_map

    @wire_map.setter
    def wire_map(self, wires):
        n = self.regularization_mesh.nC
        maps = wires.maps
        for _, mapping in maps:
            map_n = mapping.shape[0]
            if n != map_n:
                raise ValueError(
                    f"All mapping outputs must match the number of cells in "
                    f"the regularization mesh! Got {n} and {map_n}"
                )
        self._wire_map = wires

    def __call__(self, model):
        """
        Computes the sum of all joint total variation values.

        Parameters
        ----------
        model : numpy.ndarray
            stacked array of individual models np.r_[model1, model2,...]

        Returns
        -------
        float
            The computed value of the joint total variation term.
        """
        W = self.W
        G = self._G
        v2 = self.regularization_mesh.vol**2
        g2 = 0
        for m in self.wire_map * model:
            g_m = G @ m
            g2 += g_m**2
        W_g = W @ g2
        sq = np.sqrt(W_g + self.eps * v2)
        return np.sum(sq)

    def deriv(self, model):
        """
        Computes the derivative of the joint total variation.

        Parameters
        ----------
        model : numpy.ndarray
            stacked array of individual models np.r_[model1, model2,...]

        Returns
        -------
        numpy.ndarray
            The gradient of joint total variation  with respect to the model
        """
        W = self.W
        G = self._G
        g2 = 0
        gs = []
        v2 = self.regularization_mesh.vol**2
        for m in self.wire_map * model:
            g_mi = G @ m
            g2 += g_mi**2
            gs.append(g_mi)
        W_g = W @ g2
        sq = np.sqrt(W_g + self.eps * v2)
        mid = W.T @ (1 / sq)
        ps = []
        for g_mi in gs:
            ps.append(G.T @ (mid * g_mi))
        return np.concatenate(ps)

    def deriv2(self, model, v=None):
        """
        Computes the Hessian of the joint total variation.

        Parameters
        ----------
        model : numpy.ndarray
            Stacked array of individual models
        v : numpy.ndarray, optional
            An array to multiply the Hessian by.

        Returns
        -------
        numpy.ndarray or scipy.sparse.csr_matrix
            The Hessian of joint total variation with respect to the model times a
            vector or the full Hessian if `v` is `None`.
        """
        W = self.W
        G = self._G
        v2 = self.regularization_mesh.vol**2
        gs = []
        g2 = 0
        for m in self.wire_map * model:
            g_m = G @ m
            g2 += g_m**2
            gs.append(g_m)

        W_g = W @ g2
        sq = np.sqrt(W_g + self.eps * v2)
        mid = W.T @ (1 / sq)

        if v is not None:
            g_vs = []
            tmp_sum = 0
            for vi, g_i in zip(self.wire_map * v, gs):
                g_vi = G @ vi
                tmp_sum += W.T @ ((W @ (g_i * g_vi)) / sq**3)
                g_vs.append(g_vi)
            ps = []
            for g_vi, g_i in zip(g_vs, gs):
                ps.append(G.T @ (mid * g_vi - g_i * tmp_sum))
            return np.concatenate(ps)
        else:
            Pieces = []
            Diags = []
            SQ = sp.diags(sq**-1.5)
            diag_block = G.T @ sp.diags(mid) @ G
            for g_mi in gs:
                Pieces.append(SQ @ W @ sp.diags(g_mi) @ G)
                Diags.append(diag_block)
            Row = sp.hstack(Pieces, format="csr")
            Diag = sp.block_diag(Diags, format="csr")
            return Diag - Row.T @ Row



import numpy as np
from scipy.constants import epsilon_0

from ....fields import TimeFields
from ....utils import Identity, Zero, validate_type
from ....simulation import BaseSimulation


class Fields2D(TimeFields):
    r"""
    Fancy Field Storage for a 2.5D code.

    u[:,'phi', kyInd] = phi
    print(u[src0,'phi'])

    Only one field type is stored for
    each problem, the rest are computed. The fields obejct acts like an array
    and is indexed by

    .. code-block:: python

        f = problem.fields(m)
        e = f[source_list,'e']
        j = f[source_list,'j']

    If accessing all sources for a given field, use the

    .. code-block:: python

        f = problem.fields(m)
        phi = f[:,'phi']
        e = f[:,'e']
        b = f[:,'b']

    The array returned will be size (``nE`` or ``nF``, ``nSrcs`` :math:`\times`
    ``nFrequencies``)

    """

    knownFields = {}
    dtype = float

    @TimeFields.simulation.setter
    def simulation(self, value):
        self._simulation = validate_type(
            "simulation", value, BaseSimulation, cast=False
        )

    @property
    def survey(self):
        mini = self.simulation._mini_survey
        if mini is not None:
            return mini
        return self.simulation.survey

    def _phiDeriv(self, kyInd, src, du_dm_v, v, adjoint=False):
        if (
            getattr(self, "_phiDeriv_u", None) is None
            or getattr(self, "_phiDeriv_m", None) is None
        ):
            raise NotImplementedError(
                "Getting phiDerivs from {0!s} is not "
                "implemented".format(self.knownFields.keys()[0])
            )

        if adjoint:
            return (
                self._phiDeriv_u(kyInd, src, v, adjoint=adjoint),
                self._phiDeriv_m(kyInd, src, v, adjoint=adjoint),
            )

        return np.array(
            self._phiDeriv_u(kyInd, src, du_dm_v, adjoint)
            + self._phiDeriv_m(kyInd, src, v, adjoint),
            dtype=float,
        )

    def _eDeriv(self, kyInd, src, du_dm_v, v, adjoint=False):
        if (
            getattr(self, "_eDeriv_u", None) is None
            or getattr(self, "_eDeriv_m", None) is None
        ):
            raise NotImplementedError(
                "Getting eDerivs from {0!s} is not "
                "implemented".format(self.knownFields.keys()[0])
            )

        if adjoint:
            return (
                self._eDeriv_u(kyInd, src, v, adjoint),
                self._eDeriv_m(kyInd, src, v, adjoint),
            )
        return np.array(
            self._eDeriv_u(kyInd, src, du_dm_v, adjoint)
            + self._eDeriv_m(kyInd, src, v, adjoint),
            dtype=float,
        )

    def _jDeriv(self, kyInd, src, du_dm_v, v, adjoint=False):
        if (
            getattr(self, "_jDeriv_u", None) is None
            or getattr(self, "_jDeriv_m", None) is None
        ):
            raise NotImplementedError(
                "Getting jDerivs from {0!s} is not "
                "implemented".format(self.knownFields.keys()[0])
            )

        if adjoint:
            return (
                self._jDeriv_u(kyInd, src, v, adjoint),
                self._jDeriv_m(kyInd, src, v, adjoint),
            )
        return np.array(
            self._jDeriv_u(kyInd, src, du_dm_v, adjoint)
            + self._jDeriv_m(kyInd, src, v, adjoint),
            dtype=float,
        )

    def _phi_ky(self, phiSolution, source_list, kyInd):
        return phiSolution

    def _phi(self, phiSolution, source_list):
        return phiSolution.dot(self.simulation._quad_weights)

    def _phiDeriv_u(self, kyInd, src, v, adjoint=False):
        return Identity() * v

    def _phiDeriv_m(self, kyInd, src, v, adjoint=False):
        return Zero()

    def _j(self, phiSolution, source_list):
        raise NotImplementedError()

    def _e(self, phiSolution, source_list):
        raise NotImplementedError()


class Fields2DCellCentered(Fields2D):
    """
    Fancy Field Storage for a 2.5D cell centered code.
    """

    knownFields = {"phiSolution": "CC"}
    aliasFields = {
        "phi": ["phiSolution", "CC", "_phi"],
        "j": ["phiSolution", "F", "_j"],
        "e": ["phiSolution", "F", "_e"],
        "charge": ["phiSolution", "CC", "_charge"],
        "charge_density": ["phiSolution", "CC", "_charge_density"],
    }

    # primary - secondary
    # CC variables
    def _GLoc(self, fieldType):
        if fieldType == "phi":
            return "CC"
        elif fieldType == "e" or fieldType == "j":
            return "F"
        else:
            raise Exception("Field type must be phi, e, j")

    def _j(self, phiSolution, source_list):
        phi_ky = phiSolution
        sim = self.simulation
        if sim.bc_type == "Dirichlet":
            phi = self._phi(phi_ky, source_list)
            return sim.MfRhoI @ (sim.Grad @ phi)
        j = np.zeros((sim.mesh.n_faces, phi_ky.shape[1]))
        for i, (ky, w) in enumerate(zip(sim._quad_points, sim._quad_weights)):
            j += (
                sim.MfRhoI
                * (sim.Grad @ phi_ky[..., i] - sim._MBC[ky] @ phi_ky[..., i])
                * w
            )
        return j

    def _e(self, phiSolution, source_list):
        phi_ky = phiSolution
        sim = self.simulation
        if sim.bc_type == "Dirichlet":
            phi = self._phi(phi_ky, source_list)
            return sim.MfI @ (sim.Grad @ phi)
        e = np.zeros((sim.mesh.n_faces, phi_ky.shape[1]))
        for i, (ky, w) in enumerate(zip(sim._quad_points, sim._quad_weights)):
            e += (
                sim.MfI
                * (sim.Grad @ phi_ky[..., i] - sim._MBC[ky] @ phi_ky[..., i])
                * w
            )
        return e

    def _charge(self, phiSolution, source_list):
        r"""
        .. math::

            \int \nabla \codt \vec{e} =  \int \frac{\rho_v }{\epsillon_0}
        """
        sim = self.simulation
        return (
            epsilon_0
            * sim.mesh.cell_volumes[:, None]
            * (sim.mesh.face_divergence * self._e(phiSolution, source_list))
        )

    def _charge_density(self, phiSolution, source_list):
        r"""
        .. math::

            \frac{1}{V}\int \nabla \codt \vec{e} =
            \frac{1}{V}\int \frac{\rho_v }{\epsillon_0}
        """
        sim = self.simulation
        return epsilon_0 * (
            sim.mesh.face_divergence * self._e(phiSolution, source_list)
        )


class Fields2DNodal(Fields2D):
    """
    Fancy Field Storage for a 2.5D nodal code.
    """

    knownFields = {"phiSolution": "N"}
    aliasFields = {
        "phi": ["phiSolution", "N", "_phi"],
        "j": ["phiSolution", "E", "_j"],
        "e": ["phiSolution", "E", "_e"],
        "charge_density": ["phiSolution", "CC", "_charge_density"],
        "charge": ["phiSolution", "N", "_charge"],
    }
    # primary - secondary
    # CC variables

    def _GLoc(self, fieldType):
        if fieldType == "phi":
            return "N"
        elif fieldType == "e" or fieldType == "j":
            return "E"
        else:
            raise Exception("Field type must be phi, e, j")

    def _j(self, phiSolution, source_list):
        sim = self.simulation
        return sim.MeI * sim.MeSigma * self._e(phiSolution, source_list)

    def _e(self, phiSolution, source_list):
        r"""
        In HJ formulation e is not well-defined!!

        .. math::

            \vec{e} = -\nabla \phi
        """
        return -self.mesh.nodal_gradient * self._phi(phiSolution, source_list)

    def _charge(self, phiSolution, source_list):
        r"""
        .. math::

            \int \nabla \codt \vec{e} =  \int \frac{\rho_v }{\epsillon_0}
        """
        return -epsilon_0 * (
            self.mesh.nodal_gradient.T
            * self.mesh.get_edge_inner_product()
            * self._e(phiSolution, source_list)
        )

    def _charge_density(self, phiSolution, source_list):
        return (
            self.mesh.aveN2CC * self._charge(phiSolution, source_list)
        ) / self.mesh.cell_volumes[:, None]


Fields2DCellCentred = Fields2DCellCentered



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib
import warnings

from discretize import TensorMesh, TreeMesh
from discretize.base import BaseMesh
from discretize.utils import refine_tree_xyz, unpack_widths, active_from_xyz

from ....utils import (
    sdiag,
    uniqueRows,
    plot2Ddata,
    validate_type,
    validate_integer,
    validate_string,
    validate_ndarray_with_shape,
    validate_float,
)
from ..utils import geometric_factor
from . import sources as Src
from . import receivers as Rx
from .survey import Survey


class IO:
    def __init__(
        self,
        survey_layout="SURFACE",
        survey_type="dipole-dipole",
        dimension=2,
        a_locations=None,
        b_locations=None,
        m_locations=None,
        n_location=None,
        electrode_locations=None,
        data_dc_type="volt",
        data_dc=None,
        data_ip_type="volt",
        data_ip=None,
        data_sip_type="volt",
        data_sip=None,
        times_ip=None,
        G=None,
        grids=None,
        space_type="half-space",
        line_inds=None,
        sort_inds=None,
        mesh=None,
        dx=None,
        dy=None,
        dz=None,
        npad_x=5,
        npad_y=5,
        npad_z=5,
        pad_rate_x=1.3,
        pad_rate_y=1.3,
        pad_rate_z=1.3,
        ncell_per_dipole=4,
        corezlength=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.survey_layout = survey_layout
        self.survey_type = survey_type
        self.dimension = dimension
        self.a_locations = a_locations
        self.b_locations = b_locations
        self.m_locations = m_locations
        self.n_location = n_location
        self.electrode_locations = electrode_locations
        self.data_dc_type = data_dc_type
        self.data_dc = data_dc
        self.data_ip_type = data_ip_type
        self.data_ip = data_ip
        self.data_sip_type = data_sip_type
        self.data_sip = data_sip
        self.times_ip = times_ip
        self.G = G
        self.grids = grids
        self.space_type = space_type
        self.line_inds = line_inds
        self.sort_inds = sort_inds
        self.mesh = mesh
        self.dx = dx
        self.dy = dy
        self.dz = dz
        self.npad_x = npad_x
        self.npad_y = npad_y
        self.npad_z = npad_z
        self.pad_rate_x = pad_rate_x
        self.pad_rate_y = pad_rate_y
        self.pad_rate_z = pad_rate_z
        self.ncell_per_dipole = ncell_per_dipole
        self.corezlength = corezlength
        warnings.warn("code under construction - API might change in the future")

    @property
    def survey_layout(self):
        """Survey geometry of DC surveys.

        Returns
        -------
        {"SURFACE", "BOREHOLE", "GENERAL"}
        """
        return self._survey_layout

    @survey_layout.setter
    def survey_layout(self, value):
        self._survey_layout = validate_string(
            "survey_layout", value, ("SURFACE", "BOREHOLE", "GENERAL")
        )

    @property
    def survey_type(self):
        """Survey geometry of DC surveys.

        Returns
        -------
        {"dipole-dipole", "pole-dipole", "dipole-pole", "pole-pole"}
        """
        return self._survey_type

    @survey_type.setter
    def survey_type(self, value):
        self._survey_type = validate_string(
            "survey_type",
            value,
            ("dipole-dipole", "pole-dipole", "dipole-pole", "pole-pole"),
        )

    @property
    def dimension(self):
        """Dimension of electrode locations.

        Returns
        -------
        int
        """
        return self._dimension

    @dimension.setter
    def dimension(self, value):
        self._dimension = validate_integer("dimension", value, min_val=2, max_val=3)

    @property
    def a_locations(self):
        """Locations of the positive (+) current electrodes.

        Returns
        -------
        numpy.ndarray of float
        """
        return self._a_locations

    @a_locations.setter
    def a_locations(self, value):
        if value is not None:
            value = validate_ndarray_with_shape("a_locations", value, shape=("*", "*"))
        self._a_locations = value

    @property
    def b_locations(self):
        """Locations of the negative (-) current electrodes.

        Returns
        -------
        numpy.ndarray of float
        """
        return self._b_locations

    @b_locations.setter
    def b_locations(self, value):
        if value is not None:
            value = validate_ndarray_with_shape("b_locations", value, shape=("*", "*"))
        self._b_locations = value

    @property
    def m_locations(self):
        """Locations of the positive (+) potential electrodes.

        Returns
        -------
        numpy.ndarray of float
        """
        return self._m_locations

    @m_locations.setter
    def m_locations(self, value):
        if value is not None:
            value = validate_ndarray_with_shape("m_locations", value, shape=("*", "*"))
        self._m_locations = value

    @property
    def n_locations(self):
        """Locations of the negative (-) potential electrodes.

        Returns
        -------
        numpy.ndarray of float
        """
        return self._n_locations

    @n_locations.setter
    def n_locations(self, value):
        if value is not None:
            value = validate_ndarray_with_shape("n_locations", value, shape=("*", "*"))
        self._n_locations = value

    @property
    def electrode_locations(self):
        """Unique locations of a, b, m, n electrodes.

        Returns
        -------
        numpy.ndarray of float
        """
        return self._electrode_locations

    @electrode_locations.setter
    def electrode_locations(self, value):
        if value is not None:
            value = validate_ndarray_with_shape(
                "electrode_locations", value, shape=("*", "*")
            )
        self._electrode_locations = value

    # Data
    @property
    def data_dc_type(self):
        """Type of DC-IP survey.

        Returns
        -------
        {"volt", "apparent_resistivity", "apparent_conductivity"}
        """
        return self._data_dc_type

    @data_dc_type.setter
    def data_dc_type(self, value):
        self._data_dc_type = validate_string(
            "data_dc_type",
            value,
            ("volt", "apparent_resistivity", "apparent_conductivity"),
        )

    @property
    def data_dc(self):
        """Measured DC data.

        Returns
        -------
        numpy.ndarray of float
        """
        return self._data_dc

    @data_dc.setter
    def data_dc(self, value):
        if value is not None:
            value = validate_ndarray_with_shape("data_dc", value, shape=("*",))
        self._data_dc = value

    @property
    def data_ip_type(self):
        """Type of DC-IP survey.

        Returns
        -------
        {"volt", "apparent_chargeability"}
        """
        return self._data_ip_type

    @data_ip_type.setter
    def data_ip_type(self, value):
        self._data_ip_type = validate_string(
            "data_ip_type", value, ("volt", "apparent_chargeability")
        )

    @property
    def data_ip(self):
        """Measured IP data.

        Returns
        -------
        numpy.ndarray of float
        """
        return self._data_ip

    @data_ip.setter
    def data_ip(self, value):
        if value is not None:
            value = validate_ndarray_with_shape("data_ip", value, shape=("*",))
        self._data_ip = value

    @property
    def data_sip_type(self):
        """Type of DC-IP survey.

        Returns
        -------
        {"volt", "apparent_chargeability"}
        """
        return self._data_sip_type

    @data_sip_type.setter
    def data_sip_type(self, value):
        self._data_sip_type = validate_string(
            "data_sip_type", value, ("volt", "apparent_chargeability")
        )

    @property
    def data_sip(self):
        """Measured Spectral IP data.

        Returns
        -------
        numpy.ndarray of float
        """
        return self._data_sip

    @data_sip.setter
    def data_sip(self, value):
        if value is not None:
            value = validate_ndarray_with_shape("data_sip", value, shape=("*",))
        self._data_sip = value

    @property
    def times_ip(self):
        """Time channels of measured Spectral IP voltages (s).

        Returns
        -------
        numpy.ndarray of float
        """
        return self._data_sip

    @times_ip.setter
    def times_ip(self, value):
        if value is not None:
            value = validate_ndarray_with_shape("times_ip", value, shape=("*",))
        self._times_ip = value

    @property
    def G(self):
        """Geometric factor of DC-IP survey.

        Returns
        -------
        numpy.ndarray of float
        """
        return self._G

    @G.setter
    def G(self, value):
        if value is not None:
            value = validate_ndarray_with_shape("G", value, shape=("*",))
        self._G = value

    @property
    def grids(self):
        """Geometric factor of DC-IP survey.

        Returns
        -------
        numpy.ndarray of float
        """
        return self._grids

    @grids.setter
    def grids(self, value):
        if value is not None:
            value = validate_ndarray_with_shape("grids", value, shape=("*", "*"))
        self._grids = value

    @property
    def space_type(self):
        """Type of DC-IP survey.

        Returns
        -------
        {"half-space", "whole-space"}
        """
        return self._space_type

    @space_type.setter
    def space_type(self, value):
        self._space_type = validate_string(
            "space_type", value, ("half-space", "whole-space")
        )

    @property
    def line_inds(self):
        """Line indices.

        Returns
        -------
        numpy.ndarray of int
        """
        return self._line_inds

    @line_inds.setter
    def line_inds(self, value):
        if value is not None:
            value = validate_ndarray_with_shape(
                "line_inds", value, shape=("*",), dtype=int
            )
        self._line_inds = value

    @property
    def sort_inds(self):
        """Sorting indices from ABMN

        Returns
        -------
        numpy.ndarray of int
        """
        return self._sort_inds

    @sort_inds.setter
    def sort_inds(self, value):
        if value is not None:
            value = validate_ndarray_with_shape(
                "sort_inds", value, shape=("*",), dtype=int
            )
        self._sort_inds = value

    # Related to Physics and Discretization
    @property
    def mesh(self):
        """Mesh for discretization.

        Returns
        -------
        discretize.base.BaseMesh
        """
        return self._mesh

    @mesh.setter
    def mesh(self, value):
        if value is not None:
            value = validate_type("mesh", value, BaseMesh, cast=False)
        self._mesh = value

    @property
    def dx(self):
        """Length of corecell in x-direction.

        Returns
        -------
        float
        """
        return self._dx

    @dx.setter
    def dx(self, value):
        if value is not None:
            value = validate_float("dx", value, min_val=0.0, inclusive_min=False)
        self._dx = value

    @property
    def dy(self):
        """Length of corecell in y-direction.

        Returns
        -------
        float
        """
        return self._dy

    @dy.setter
    def dy(self, value):
        if value is not None:
            value = validate_float("dy", value, min_val=0.0, inclusive_min=False)
        self._dy = value

    @property
    def dz(self):
        """Length of corecell in z-direction.

        Returns
        -------
        float
        """
        return self._dz

    @dz.setter
    def dz(self, value):
        if value is not None:
            value = validate_float("dz", value, min_val=0.0, inclusive_min=False)
        self._dz = value

    @property
    def npad_x(self):
        """The number of padding cells x-direction.

        Returns
        -------
        int
        """
        return self._npad_x

    @npad_x.setter
    def npad_x(self, value):
        self._npad_x = validate_integer("npad_x", value, min_val=0)

    @property
    def npad_y(self):
        """The number of padding cells y-direction.

        Returns
        -------
        int
        """
        return self._npad_y

    @npad_y.setter
    def npad_y(self, value):
        self._npad_y = validate_integer("npad_y", value, min_val=0)

    @property
    def npad_z(self):
        """The number of padding cells y-direction.

        Returns
        -------
        int
        """
        return self._npad_z

    @npad_z.setter
    def npad_z(self, value):
        self._npad_z = validate_integer("npad_z", value, min_val=0)

    @property
    def pad_rate_x(self):
        """Expansion rate of padding cells in  x-direction.

        Returns
        -------
        float
        """
        return self._pad_rate_x

    @pad_rate_x.setter
    def pad_rate_x(self, value):
        self._pad_rate_x = validate_float("pad_rate_x", value)

    @property
    def pad_rate_y(self):
        """Expansion rate of padding cells in  x-direction.

        Returns
        -------
        float
        """
        return self._pad_rate_y

    @pad_rate_y.setter
    def pad_rate_y(self, value):
        self._pad_rate_y = validate_float("pad_rate_y", value)

    @property
    def pad_rate_z(self):
        """Expansion rate of padding cells in  x-direction.

        Returns
        -------
        float
        """
        return self._pad_rate_z

    @pad_rate_z.setter
    def pad_rate_z(self, value):
        self._pad_rate_z = validate_float("pad_rate_z", value)

    @property
    def ncell_per_dipole(self):
        """The number of cells between dipole electrodes.

        Returns
        -------
        int
        """
        return self._ncell_per_dipole

    @ncell_per_dipole.setter
    def ncell_per_dipole(self, value):
        self._ncell_per_dipole = validate_integer("ncell_per_dipole", value, min_val=0)

    @property
    def corezlength(self):
        """Core depth (m).

        Returns
        -------
        float
        """
        return self._pad_rate_z

    @corezlength.setter
    def corezlength(self, value):
        if value is not None:
            value = validate_float("corezlength", value)
        self._corezlength = value

    # For synthetic surveys
    x0 = None
    lineLength = None
    a = None
    n_spacing = None
    n_data = None

    # Properties
    @property
    def voltages(self):
        """
        Votages (V)
        """
        if self.data_dc_type.lower() == "volt":
            return self.data_dc
        elif self.data_dc_type.lower() == "apparent_resistivity":
            return self.data_dc * self.G
        elif self.data_dc_type.lower() == "apparent_conductivity":
            return self.apparent_conductivity / (self.data_dc * self.G)
        else:
            raise NotImplementedError()

    @property
    def apparent_resistivity(self):
        """
        Apparent Resistivity (Ohm-m)
        """
        if self.data_dc_type.lower() == "apparent_resistivity":
            return self.data_dc
        elif self.data_dc_type.lower() == "volt":
            return self.data_dc / self.G
        elif self.data_dc_type.lower() == "apparent_conductivity":
            return 1.0 / self.data_dc
        else:
            print(self.data_dc_type.lower())
            raise NotImplementedError()

    @property
    def apparent_conductivity(self):
        """
        Apparent Conductivity (S/m)
        """
        if self.data_dc_type.lower() == "apparent_conductivity":
            return self.data_dc
        elif self.data_dc_type.lower() == "apparent_resistivity":
            return 1.0 / self.data_dc
        elif self.data_dc_type.lower() == "volt":
            return 1.0 / self.data_dc * self.G

    # For IP
    @property
    def voltages_ip(self):
        """
        IP votages (V)
        """
        if self.data_ip_type.lower() == "volt":
            return self.data_ip
        elif self.data_ip_type.lower() == "apparent_chargeability":
            if self.voltages is None:
                raise Exception("DC voltages must be set to compute IP voltages")
            return self.data_ip * self.voltages
        else:
            raise NotImplementedError()

    # For SIP
    @property
    def voltages_sip(self):
        """
        IP votages (V)
        """
        if self.data_sip_type.lower() == "volt":
            return self.data_sip
        elif self.data_sip_type.lower() == "apparent_chargeability":
            if self.voltages is None:
                raise Exception("DC voltages must be set to compute IP voltages")
            return sdiag(self.voltages) * self.data_sip
        else:
            raise NotImplementedError()

    @property
    def apparent_chargeability(self):
        """
        Apparent Conductivity (S/m)
        """
        if self.data_ip_type.lower() == "apparent_chargeability":
            return self.data_ip
        elif self.data_ip_type.lower() == "volt":
            if self.voltages is None:
                raise Exception(
                    "DC voltages must be set to compute Apparent Chargeability"
                )
            return self.data_ip / self.voltages
        else:
            raise NotImplementedError()

    # For SIP
    @property
    def apparent_chargeability_sip(self):
        """
        Apparent Conductivity (S/m)
        """
        if self.data_sip_type.lower() == "apparent_chargeability":
            return self.data_sip
        elif self.data_sip_type.lower() == "volt":
            if self.voltages is None:
                raise Exception(
                    "DC voltages must be set to compute Apparent Chargeability"
                )
            return sdiag(1.0 / self.voltages) * self.data_sip
        else:
            raise NotImplementedError()

    def geometric_factor(self, survey):
        """
        Compute geometric factor, G, using locational informaition
        in survey object
        """
        G = geometric_factor(survey, space_type=self.space_type)
        return G

    def from_ambn_locations_to_survey(self, *args, **kwargs):
        raise NotImplementedError(
            "from_ambn_locations_to_survey has been renamed to "
            "from_abmn_locations_to_survey. It will be removed in a future version 0.17.0 of simpeg",
        )

    def from_abmn_locations_to_survey(
        self,
        a_locations,
        b_locations,
        m_locations,
        n_locations,
        survey_type=None,
        data_dc=None,
        data_ip=None,
        data_sip=None,
        data_dc_type="volt",
        data_ip_type="volt",
        data_sip_type="volt",
        fname=None,
        dimension=2,
        line_inds=None,
        times_ip=None,
    ):
        """
        read A, B, M, N electrode location and data (V or apparent_resistivity)
        """
        self.a_locations = a_locations.copy()
        self.b_locations = b_locations.copy()
        self.m_locations = m_locations.copy()
        self.n_locations = n_locations.copy()
        self.survey_type = survey_type
        self.dimension = dimension
        self.data_dc_type = data_dc_type
        self.data_ip_type = data_ip_type
        self.data_sip_type = data_sip_type
        if times_ip is not None:
            self.times_ip = times_ip

        uniqSrc = uniqueRows(np.c_[self.a_locations, self.b_locations])
        uniqElec = uniqueRows(
            np.vstack(
                (self.a_locations, self.b_locations, self.m_locations, self.n_locations)
            )
        )
        self.electrode_locations = uniqElec[0]
        nSrc = uniqSrc[0].shape[0]
        ndata = self.a_locations.shape[0]

        if self.survey_layout == "SURFACE":
            # 2D locations
            source_lists = []
            sort_inds = []
            for iSrc in range(nSrc):
                inds = uniqSrc[2] == iSrc
                sort_inds.append(np.arange(ndata)[inds])

                locsM = self.m_locations[inds, :]
                locsN = self.n_locations[inds, :]
                if survey_type in ["dipole-dipole", "pole-dipole"]:
                    rx = Rx.Dipole(locsM, locsN)
                elif survey_type in ["dipole-pole", "pole-pole"]:
                    rx = Rx.Pole(locsM)
                else:
                    raise NotImplementedError()

                if dimension == 2:
                    locA = uniqSrc[0][iSrc, :2]
                    locB = uniqSrc[0][iSrc, 2:]
                elif dimension == 3:
                    locA = uniqSrc[0][iSrc, :3]
                    locB = uniqSrc[0][iSrc, 3:]

                if survey_type in ["dipole-dipole", "dipole-pole"]:
                    src = Src.Dipole([rx], locA, locB)
                elif survey_type in ["pole-dipole", "pole-pole"]:
                    src = Src.Pole([rx], locA)

                source_lists.append(src)

            self.sort_inds = np.hstack(sort_inds)

            if dimension in (2, 3):
                survey = Survey(source_lists)
            else:
                raise NotImplementedError()

            self.a_locations = self.a_locations[self.sort_inds, :]
            self.b_locations = self.b_locations[self.sort_inds, :]
            self.m_locations = self.m_locations[self.sort_inds, :]
            self.n_locations = self.n_locations[self.sort_inds, :]
            self.G = self.geometric_factor(survey)

            if data_dc is not None:
                self.data_dc = data_dc[self.sort_inds]
            if data_ip is not None:
                self.data_ip = data_ip[self.sort_inds]
            if data_sip is not None:
                self.data_sip = data_sip[self.sort_inds, :]
            if line_inds is not None:
                self.line_inds = line_inds[self.sort_inds]
            # Here we ignore ... z-locations
            self.n_data = survey.nD

            midABx = (self.a_locations[:, 0] + self.b_locations[:, 0]) * 0.5
            midMNx = (self.m_locations[:, 0] + self.n_locations[:, 0]) * 0.5

            if dimension == 2:
                z = abs(midABx - midMNx) * 1.0 / 3.0
                x = (midABx + midMNx) * 0.5
                zmax = z.max()
                a = abs(np.diff(np.sort(self.electrode_locations[:, 0]))).min()
                # Consider the case of 1D types of array
                if np.all(zmax < a):
                    z = abs(self.a_locations[:, 0] - self.b_locations[:, 0]) / 3.0
                self.grids = np.c_[x, z]

            elif dimension == 3:
                midABy = (self.a_locations[:, 1] + self.b_locations[:, 1]) * 0.5
                midMNy = (self.m_locations[:, 1] + self.n_locations[:, 1]) * 0.5
                z = np.sqrt((midABx - midMNx) ** 2 + (midABy - midMNy) ** 2) * 1.0 / 3.0
                x = (midABx + midMNx) * 0.5
                y = (midABy + midMNy) * 0.5
                self.grids = np.c_[x, y, z]
            else:
                raise Exception()
        else:
            raise NotImplementedError()
        return survey

    def set_mesh(
        self,
        topo=None,
        dx=None,
        dy=None,
        dz=None,
        corezlength=None,
        npad_x=None,
        npad_y=None,
        npad_z=None,
        pad_rate_x=None,
        pad_rate_y=None,
        pad_rate_z=None,
        ncell_per_dipole=None,
        mesh_type="TensorMesh",
        dimension=2,
        method="nearest",
    ):
        """
        Set up a mesh for a given DC survey
        """

        # Update properties
        if npad_x is None:
            npad_x = self.npad_x
        self.npad_x = npad_x

        if npad_z is None:
            npad_z = self.npad_z
        self.npad_z = npad_z

        if pad_rate_x is None:
            pad_rate_x = self.pad_rate_x
        self.pad_rate_x = pad_rate_x

        if pad_rate_z is None:
            pad_rate_z = self.pad_rate_z
        self.pad_rate_z = pad_rate_z

        if ncell_per_dipole is None:
            ncell_per_dipole = self.ncell_per_dipole
        self.ncell_per_dipole = ncell_per_dipole

        # 2D or 3D mesh
        if dimension not in [2, 3]:
            raise NotImplementedError(
                "Set mesh has not been implemented for a 1D system"
            )

        if dimension == 2:
            z_ind = 1
        else:
            z_ind = 2
        a = abs(np.diff(np.sort(self.electrode_locations[:, 0]))).min()
        lineLength = abs(
            self.electrode_locations[:, 0].max() - self.electrode_locations[:, 0].min()
        )
        dx_ideal = a / ncell_per_dipole
        if dx is None:
            dx = dx_ideal
            print(
                "dx is set to {} m (samllest electrode spacing ({}) / {})".format(
                    dx, a, ncell_per_dipole
                )
            )
        if dz is None:
            dz = dx * 0.5
            print("dz ({} m) is set to dx ({} m) / {}".format(dz, dx, 2))
        if dimension == 3:
            if dy is None:
                print("dy is set equal to dx")
                dy = dx
            self.dy = dy

            if npad_y is None:
                npad_y = self.npad_y
            self.npad_y = npad_y

            if pad_rate_y is None:
                pad_rate_y = self.pad_rate_y
            self.pad_rate_y = pad_rate_y

        x0 = self.electrode_locations[:, 0].min()
        if topo is None:
            # For 2D mesh
            if dimension == 2:
                # sort by x
                row_idx = np.lexsort((self.electrode_locations[:, 0],))
            # For 3D mesh
            else:
                # sort by x, then by y
                row_idx = np.lexsort(
                    (self.electrode_locations[:, 1], self.electrode_locations[:, 0])
                )
            locs = self.electrode_locations[row_idx, :]
        else:
            # For 2D mesh
            if dimension == 2:
                mask = np.isin(self.electrode_locations[:, 0], topo[:, 0])
                if np.any(mask):
                    warnings.warn(
                        "Because the x coordinates of some topo and electrodes are the same,"
                        " we excluded electrodes with the same coordinates.",
                        RuntimeWarning,
                    )
                locs_tmp = np.vstack((topo, self.electrode_locations[~mask, :]))
                row_idx = np.lexsort((locs_tmp[:, 0],))
            else:
                dtype = [("x", np.float64), ("y", np.float64)]
                mask = np.isin(
                    self.electrode_locations[:, [0, 1]].copy().view(dtype),
                    topo[:, [0, 1]].copy().view(dtype),
                ).flatten()
                if np.any(mask):
                    warnings.warn(
                        "Because the x and y coordinates of some topo and electrodes are the same,"
                        " we excluded electrodes with the same coordinates.",
                        RuntimeWarning,
                    )
                locs_tmp = np.vstack((topo, self.electrode_locations[~mask, :]))
                row_idx = np.lexsort((locs_tmp[:, 1], locs_tmp[:, 0]))
            locs = locs_tmp[row_idx, :]

        if dx > dx_ideal:
            # warnings.warn(
            #     "Input dx ({}) is greater than expected \n We recommend using {:0.1e} m cells, that is, {} cells per {0.1e} m dipole length".format(dx, dx_ideal, ncell_per_dipole, a)
            # )
            pass

        # Set mesh properties to class instance
        self.dx = dx
        self.dz = dz

        zmax = locs[:, z_ind].max()

        # 3 cells each for buffer
        corexlength = lineLength + dx * 6
        if corezlength is None:
            dz_topo = locs[:, 1].max() - locs[:, 1].min()
            corezlength = self.grids[:, z_ind].max() + dz_topo
            self.corezlength = corezlength

        if mesh_type == "TensorMesh":
            ncx = np.round(corexlength / dx)
            ncz = np.round(corezlength / dz)
            hx = [(dx, npad_x, -pad_rate_x), (dx, ncx), (dx, npad_x, pad_rate_x)]
            hz = [(dz, npad_z, -pad_rate_z), (dz, ncz)]
            x0_mesh = -(
                (dx * pad_rate_x ** (np.arange(npad_x) + 1)).sum() + dx * 3 - x0
            )
            z0_mesh = (
                -((dz * pad_rate_z ** (np.arange(npad_z) + 1)).sum() + dz * ncz) + zmax
            )

            # For 2D mesh
            if dimension == 2:
                h = [hx, hz]
                x0_for_mesh = [x0_mesh, z0_mesh]
                self.xyzlim = np.vstack(
                    (np.r_[x0, x0 + lineLength], np.r_[zmax - corezlength, zmax])
                )

            # For 3D mesh
            else:
                ylocs = np.unique(self.electrode_locations[:, 1])
                ymin, ymax = ylocs.min(), ylocs.max()
                # 3 cells each for buffer in y-direction
                coreylength = ymax - ymin + dy * 6
                ncy = np.round(coreylength / dy)
                hy = [(dy, npad_y, -pad_rate_y), (dy, ncy), (dy, npad_y, pad_rate_y)]
                y0 = ylocs.min() - dy / 2.0
                y0_mesh = -(
                    (dy * pad_rate_y ** (np.arange(npad_y) + 1)).sum() + dy * 3 - y0
                )

                h = [hx, hy, hz]
                x0_for_mesh = [x0_mesh, y0_mesh, z0_mesh]
                self.xyzlim = np.vstack(
                    (
                        np.r_[x0, x0 + lineLength],
                        np.r_[ymin - dy * 3, ymax + dy * 3],
                        np.r_[zmax - corezlength, zmax],
                    )
                )
            mesh = TensorMesh(h, x0=x0_for_mesh)

        elif mesh_type == "TREE":
            # Quadtree mesh
            if dimension == 2:
                pad_length_x = np.sum(unpack_widths([(dx, npad_x, pad_rate_x)]))
                pad_length_z = np.sum(unpack_widths([(dz, npad_z, pad_rate_z)]))

                dom_width_x = lineLength + 2 * pad_length_x  # domain width x
                dom_width_z = corezlength + pad_length_z  # domain width z

                nbcx = 2 ** int(
                    np.ceil(np.log(dom_width_x / dx) / np.log(2.0))
                )  # num. base cells x
                nbcz = 2 ** int(
                    np.ceil(np.log(dom_width_z / dz) / np.log(2.0))
                )  # num. base cells z

                length = 0.0
                dz_tmp = dz
                octree_levels = []
                while length < corezlength:
                    length += 5 * dz_tmp
                    octree_levels.append(5)
                    dz_tmp *= 2

                # Define the base mesh
                hx = [(dx, nbcx)]
                hz = [(dz, nbcz)]

                mesh_width = np.sum(unpack_widths(hx))
                mesh_height = np.sum(unpack_widths(hz))

                array_midpoint = 0.5 * (
                    self.electrode_locations[:, 0].min()
                    + self.electrode_locations[:, 0].max()
                )
                mesh = TreeMesh(
                    [hx, hz], x0=[array_midpoint - mesh_width / 2, zmax - mesh_height]
                )
                # mesh = TreeMesh([hx, hz], x0='CN')

                # Mesh refinement based on topography
                mesh = refine_tree_xyz(
                    mesh,
                    self.electrode_locations,
                    octree_levels=octree_levels,
                    method="radial",
                    finalize=False,
                )
                mesh.finalize()

                self.xyzlim = np.vstack(
                    (
                        np.r_[
                            self.electrode_locations[:, 0].min(),
                            self.electrode_locations[:, 0].max(),
                        ],
                        np.r_[zmax - corezlength, zmax],
                    )
                )

            # Octree mesh
            elif dimension == 3:
                raise NotImplementedError(
                    "set_mesh has not implemented 3D TreeMesh (yet)"
                )

        else:
            raise NotImplementedError(
                "set_mesh currently generates TensorMesh or TreeMesh"
            )

        actind = active_from_xyz(mesh, locs, method=method)

        return mesh, actind

    def plotPseudoSection(
        self,
        data_type="apparent_resistivity",
        data=None,
        dataloc=True,
        aspect_ratio=2,
        scale="log",
        cmap="viridis",
        ncontour=10,
        ax=None,
        figname=None,
        clim=None,
        label=None,
        iline=0,
        orientation="vertical",
    ):
        """
        Plot 2D pseudo-section for DC-IP data
        """
        matplotlib.rcParams["font.size"] = 12

        if ax is None:
            fig = plt.figure(figsize=(10, 5))
            ax = plt.subplot(111)

        if self.dimension == 2:
            inds = np.ones(self.n_data, dtype=bool)
            grids = self.grids.copy()
        elif self.dimension == 3:
            inds = self.line_inds == iline
            grids = self.grids[inds, :][:, [0, 2]]
        else:
            raise NotImplementedError()

        if data_type == "apparent_resistivity":
            if data is None:
                val = self.apparent_resistivity[inds]
            else:
                val = data.copy()[inds]
            label_tmp = r"Apparent Res. ($\Omega$m)"
        elif data_type == "volt":
            if data is None:
                val = self.voltages[inds]
            else:
                val = data.copy()[inds]
            label_tmp = "Voltage (V)"
        elif data_type == "apparent_conductivity":
            if data is None:
                val = self.apparent_conductivity[inds]
            else:
                val = data.copy()[inds]
            label_tmp = "Apparent Cond. (S/m)"
        elif data_type == "apparent_chargeability":
            if data is not None:
                val = data.copy()[inds]
            else:
                val = self.apparent_chargeability.copy()[inds] * 1e3
            label_tmp = "Apparent Charg. (mV/V)"
        elif data_type == "volt_ip":
            if data is not None:
                val = data.copy()[inds]
            else:
                val = self.voltages_ip.copy()[inds] * 1e3
            label_tmp = "Secondary voltage. (mV)"
        else:
            print(data_type)
            raise NotImplementedError()

        if label is None:
            label = label_tmp

        out = plot2Ddata(
            grids,
            val,
            contourOpts={"cmap": cmap},
            ax=ax,
            dataloc=dataloc,
            scale=scale,
            ncontour=ncontour,
            clim=clim,
        )
        ax.invert_yaxis()
        ax.set_xlabel("x (m)")
        ax.set_yticklabels([])
        ax.set_ylabel("n-spacing")
        if orientation == "vertical":
            frac = 0.01
        elif orientation == "horizontal":
            frac = 0.03
        else:
            raise ValueError(
                "Orientation must be either vertical or horizontal, not {}".format(
                    orientation
                )
            )
        cb = plt.colorbar(
            out[0], format="%.1e", ax=ax, orientation=orientation, fraction=frac
        )
        cb.set_label(label)
        cb.set_ticks(out[0].levels)
        ax.set_aspect(aspect_ratio)
        plt.tight_layout()
        if figname is not None:
            fig.savefig(figname, dpi=200)

    def read_ubc_dc2d_obs_file(self, filename, input_type="simple", toponame=None):
        obsfile = np.genfromtxt(filename, delimiter=" \n", dtype=str, comments="!")
        if input_type == "general":
            topo = None
            n_src = 0
            n_rxs = []
            src_info = []
            abmn = []
            for obs in obsfile:
                temp = np.fromstring(obs, dtype=float, sep=" ").T
                if len(temp) == 5:
                    n_src += 1
                    src_info = temp[:4]
                    n_rxs.append(int(temp[-1]))
                else:
                    abmn.append(np.r_[src_info, temp])

            abmn = np.vstack(abmn)
            a = np.c_[abmn[:, 0], -abmn[:, 1]]
            b = np.c_[abmn[:, 2], -abmn[:, 3]]
            m = np.c_[abmn[:, 4], -abmn[:, 5]]
            n = np.c_[abmn[:, 6], -abmn[:, 7]]
            voltage = abmn[:, 8]
            standard_deviation = abmn[:, 9]

        elif input_type == "simple":
            if toponame is not None:
                tmp_topo = np.loadtxt(toponame)
                n_topo = tmp_topo[0, 0]
                topo = tmp_topo[1:, :]
                if topo.shape[0] != n_topo:
                    print(
                        ">> # of points for the topography is not {0}, but {0}".format(
                            n_topo, topo.shape[0]
                        )
                    )
            tmp = np.loadtxt(filename, comments="!").astype(float)
            e = np.zeros(tmp.shape[0], dtype=float)
            a = np.c_[tmp[:, 0], e]
            b = np.c_[tmp[:, 1], e]
            m = np.c_[tmp[:, 2], e]
            n = np.c_[tmp[:, 3], e]
            voltage = tmp[:, 4]
            standard_deviation = tmp[:, 5]

        if np.all(a == b):
            if np.all(m == n):
                survey_type = "pole-pole"
            else:
                survey_type = "pole-dipole"
        else:
            if np.all(m == n):
                survey_type = "dipole-pole"
            else:
                survey_type = "dipole-dipole"

        survey = self.from_abmn_locations_to_survey(
            a, b, m, n, survey_type=survey_type, data_dc=voltage
        )
        survey.dobs = voltage[self.sort_inds]
        survey.std = standard_deviation[self.sort_inds]
        survey.topo = topo
        return survey

    def write_to_csv(self, fname, dobs, standard_deviation=None, **kwargs):
        uncert = kwargs.pop("uncertainty", None)
        if uncert is not None:
            raise TypeError(
                "The uncertainty option has been removed, please use standard_deviation."
            )

        if standard_deviation is None:
            standard_deviation = np.ones(dobs.size) * np.nan
        data = np.c_[
            self.a_locations,
            self.b_locations,
            self.m_locations,
            self.n_locations,
            dobs,
            standard_deviation,
        ]
        df = pd.DataFrame(
            data=data,
            columns=[
                "Ax",
                "Az",
                "Bx",
                "Bz",
                "Mx",
                "Mz",
                "Nx",
                "Nz",
                "Voltage",
                "Uncertainty",
            ],
        )
        df.to_csv(fname)

    def read_dc_data_csv(self, fname, dim=2):
        df = pd.read_csv(fname)
        if dim == 2:
            a_locations = df[["Ax", "Az"]].values
            b_locations = df[["Bx", "Bz"]].values
            m_locations = df[["Mx", "Mz"]].values
            n_locations = df[["Nx", "Nz"]].values
            dobs = df["Voltage"].values
            standard_deviation = df["Uncertainty"].values

            if np.all(a_locations == b_locations):
                src_type = "pole-"
            else:
                src_type = "dipole-"

            if np.all(m_locations == n_locations):
                rx_type = "pole"
            else:
                rx_type = "dipole"
            survey_type = src_type + rx_type
            survey = self.from_abmn_locations_to_survey(
                a_locations,
                b_locations,
                m_locations,
                n_locations,
                survey_type,
                data_dc=dobs,
                data_dc_type="volt",
            )
            survey.std = standard_deviation[self.sort_inds]
            survey.dobs = dobs[self.sort_inds]
        else:
            raise NotImplementedError()
        return survey

    def read_topo_csv(self, fname, dim=2):
        if dim == 2:
            df = pd.read_csv(fname)
            topo = df[["X", "Z"]].values
        return topo
